#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

##############################################################################
## Parts of this code were adapted from:
##
##  -   TESS
##
##          https://github.com/hoehna/TESS
##
##          H{"o}hna S. 2013. Fast simulation of reconstructed phylogenies under
##          global time-dependent birth--death processes. Bioinformatics, 29(11)
##          1367-1374.
##
##          Copyright (c) 2012- Sebastian Hoehna
##
##          TESS is free software; you can redistribute it and/or modify
##          it under the terms of the GNU Lesser General Public License as
##          published by the Free Software Foundation; either version 2
##          of the License, or (at your option) any later version.
##
##############################################################################

"""
Models, modeling and model-fitting of birth-death processes.

-   Nee, S.  2001.  Inferring speciation rates from phylogenies.
    Evolution 55:661-668.
-   Yule, G. U. 1924. A mathematical theory of evolution based on the
    conclusions of Dr.  J. C. Willis. Phil. Trans. R. Soc. Lond. B
    213:21-87.
-   Hoehna, S. (2015). The time-dependent reconstructed evolutionary process
    with a key-role for mass-extinction events. Journal of theoretical biology,
    380, 321-331.

"""

import sys
import math
import collections
import itertools
from dendropy.calculate import combinatorics
from dendropy.calculate import probability
from dendropy.utility import GLOBAL_RNG
from dendropy.utility.error import TreeSimTotalExtinctionException
from dendropy.utility import constants
from dendropy.utility import deprecate

import dendropy

def birth_death_tree(birth_rate, death_rate, birth_rate_sd=0.0, death_rate_sd=0.0, **kwargs):
    """
    Returns a birth-death tree with birth rate specified by ``birth_rate``, and
    death rate specified by ``death_rate``, with edge lengths in continuous (real)
    units.

    Tree growth is controlled by one or more of the following arguments, of which
    at least one must be specified:

        - If ``num_extant_tips`` is given as a keyword argument, tree is grown until the
          number of EXTANT tips equals this number.
        - If ``num_extinct_tips`` is given as a keyword argument, tree is grown until the
          number of EXTINCT tips equals this number.
        - If ``num_total_tips`` is given as a keyword argument, tree is grown until the
          number of EXTANT plus EXTINCT tips equals this number.
        - If 'max_time' is given as a keyword argument, tree is grown for
          a maximum of ``max_time``.
        - If ``gsa_ntax`` is given then the tree will be simulated up to this number of
          EXTANT tips (or 0 tips), then a tree will be randomly selected from the
          intervals which corresond to times at which the tree had exactly ``num_extant_tips``
          leaves. This allows for simulations according to the "General
          Sampling Approach" of Hartmann et al. (2010). If this option is
          specified, then ``num_extant_tips`` MUST be specified and
          ``num_extinct_tips`` and ``num_total_tips`` CANNOT be specified.

    If more than one of the above is given, then tree growth will terminate when
    *any* one of the termination conditions are met.

    Parameters
    ----------

    birth_rate : float
        The birth rate.
    death_rate : float
        The death rate.
    birth_rate_sd : float
        The standard deviation of the normally-distributed mutation added to
        the birth rate as it is inherited by daughter nodes; if 0, birth rate
        does not evolve on the tree.
    death_rate_sd : float
        The standard deviation of the normally-distributed mutation added to
        the death rate as it is inherited by daughter nodes; if 0, death rate
        does not evolve on the tree.

    Keyword Arguments
    -----------------

    num_extant_tips: int
        If specified, branching process is terminated when number of EXTANT
        tips equals this number.
    num_extinct_tips: int
        If specified, branching process is terminated when number of EXTINCT
        tips equals this number.
    num_total_tips: int
        If specified, branching process is terminated when number of EXTINCT
        plus EXTANT tips equals this number.
    max_time: float
        If specified, branching process is terminated when time reaches or
        exceeds this value.
    gsa_ntax: int
        The General Sampling Approach threshold for number of taxa. See above
        for details.
    tree : Tree instance
        If given, then this tree will be used; otherwise a new one will be created.
    taxon_namespace : TaxonNamespace instance
        If given, then this will be assigned to the new tree, and, in addition,
        taxa assigned to tips will be sourced from or otherwise created with
        reference to this.
    is_assign_extant_taxa : bool [default: True]
        If False, then taxa will not be assigned to extant tips. If True
        (default), then taxa will be assigned to extant tips. Taxa will be
        assigned from the specified ``taxon_namespace`` or
        ``tree.taxon_namespace``. If the number of taxa required exceeds the
        number of taxa existing in the taxon namespace, new |Taxon| objects
        will be created as needed and added to the taxon namespace.
    is_assign_extinct_taxa : bool [default: True]
        If False, then taxa will not be assigned to extant tips. If True
        (default), then taxa will be assigned to extant tips. Taxa will be
        assigned from the specified ``taxon_namespace`` or
        ``tree.taxon_namespace``. If the number of taxa required exceeds the
        number of taxa existing in the taxon namespace, new |Taxon| objects
        will be created as needed and added to the taxon namespace. Note that
        this option only makes sense if extinct tips are retained (specified via
        'is_retain_extinct_tips' option), and will otherwise be ignored.
    is_add_extinct_attr: bool [default: True]
        If True (default), add an boolean attribute indicating whether or not a
        node is an extinct tip or not. False will skip this. Name of attribute
        is set by 'extinct_attr_name' argument, defaulting to 'is_extinct'.
        Note that this option only makes sense if extinct tips are retained
        (specified via 'is_retain_extinct_tips' option), and will otherwise be
        ignored.
    extinct_attr_name: str [default: 'is_extinct']
        Name of attribute to add to nodes indicating whether or not tip is extinct.
        Note that this option only makes sense if extinct tips are retained
        (specified via 'is_retain_extinct_tips' option), and will otherwise be
        ignored.
    is_retain_extinct_tips : bool [default: False]
        If True, extinct tips will be retained on tree. Defaults to False:
        extinct lineages removed from tree.
    repeat_until_success: bool [default: True]
        Under some conditions, it is possible for all lineages on a tree to go
        extinct. In this case, if this argument is given as |True| (the
        default), then a new branching process is initiated. If |False|
        (default), then a TreeSimTotalExtinctionException is raised.
    rng: random.Random() or equivalent instance
        A Random() object or equivalent can be passed using the ``rng`` keyword;
        otherwise GLOBAL_RNG is used.

    References
    ----------

    Hartmann, Wong, and Stadler "Sampling Trees from Evolutionary Models" Systematic Biology. 2010. 59(4). 465-476

    """
    if "assign_taxa" in kwargs:
        deprecate.dendropy_deprecation_warning(
                message="Deprecated: 'assign_taxa' will no longer be supported as an argument to this function. Use 'is_assign_extant_taxa' and/or 'is_assign_extinct_taxa' instead",
                stacklevel=3)
        a = kwargs.pop("assign_taxa")
        kwargs["is_assign_extant_taxa"] = a
        kwargs["is_assign_extant_taxa"] = a
    if "ntax" in kwargs:
        deprecate.dendropy_deprecation_warning(
                message="Deprecated: 'ntax' is no longer supported as an argument to this function. Use one or more of the following instead: 'num_extant_tips', 'num_extinct_tips', 'num_total_tips', or 'max_time'",
                stacklevel=3)
        kwargs["num_extant_tips"] = kwargs.pop("ntax")
    if (("num_extant_tips" not in kwargs)
            and ("num_extinct_tips" not in kwargs)
            and ("num_total_tips" not in kwargs)
            and ("max_time" not in kwargs) ):
        if "taxon_namespace" in kwargs:
            ### cannot support legacy approach, b/c ``taxon_namespace`` may grow during function, leading to unpredictable behavior
            # deprecate.dendropy_deprecation_warning(
            #         preamble="Deprecated: The 'taxon_namespace' argument can no longer be used to specify a termination condition as a side-effect. Use one or more of the following instead with the length of the taxon namespace instance as a value: 'num_extant_tips', 'num_extinct_tips', or 'num_total_tips'",
            #         old_construct="tree = birth_death_tree(\n    ...\n    taxon_namespace=taxon_namespace,\n    ...\n)",
            #         new_construct="tree = birth_death_tree(\n    ...\n    taxon_namespace=taxon_namespace,\n    num_extant_tips=len(taxon_namespace),\n    ...\n)")
            # kwargs["num_extant_tips"] = len(kwargs["taxon_namespace"])
            raise ValueError("The 'taxon_namespace' argument can no longer be used to specify a termination condition as a side-effect."
                             "Use one or more of the following instead with the length of the taxon namespace instance as a value: "
                             "'num_extant_tips', 'num_extinct_tips', or 'num_total_tips'.\n"
                             "That is, instead of:\n\n"
                             "    tree = birth_death_tree(\n        ...\n        taxon_namespace=taxon_namespace,\n        ...\n    )\n\n"
                             "Use:\n\n"
                             "    ntax = len(taxon_namespace)\n    tree = birth_death_tree(\n        ...\n        taxon_namespace=taxon_namespace,\n        num_extant_tips=ntax,\n        ...\n    )\n"
                             "\nOr (recommended):\n\n"
                             "    tree = birth_death_tree(\n        ...\n        taxon_namespace=taxon_namespace,\n        num_extant_tips=100,\n        ...\n    )\n"
                             "\nNote that the taxon namespace instance size may grow during any particular call of the function depending on taxon assignment/creation settings, so"
                             " for stable and predictable behavor it is important to take a snapshot of the desired taxon namespace size before any call of the function, or, better yet"
                             " simply pass in a constant value."
                             )
        else:
            raise ValueError("One or more of the following must be specified: 'num_extant_tips', 'num_extinct_tips', or 'max_time'")
    target_num_extant_tips = kwargs.pop("num_extant_tips", None)
    target_num_extinct_tips = kwargs.pop("num_extinct_tips", None)
    target_num_total_tips = kwargs.pop("num_total_tips", None)
    max_time = kwargs.pop('max_time', None)
    gsa_ntax = kwargs.pop('gsa_ntax', None)
    is_add_extinct_attr = kwargs.pop('is_add_extinct_attr', True)
    extinct_attr_name = kwargs.pop('extinct_attr_name', 'is_extinct')
    is_retain_extinct_tips = kwargs.pop('is_retain_extinct_tips', False)
    is_assign_extant_taxa = kwargs.pop('is_assign_extant_taxa', True)
    is_assign_extinct_taxa = kwargs.pop('is_assign_extinct_taxa', True)
    repeat_until_success = kwargs.pop('repeat_until_success', True)

    tree = kwargs.pop("tree", None)
    taxon_namespace = kwargs.pop("taxon_namespace", None)

    rng = kwargs.pop('rng', GLOBAL_RNG)

    ignore_unrecognized_keyword_arguments = kwargs.pop('ignore_unrecognized_keyword_arguments', False)
    if kwargs and not ignore_unrecognized_keyword_arguments:
        raise ValueError("Unsupported keyword arguments: {}".format(kwargs.keys()))

    terminate_at_full_tree = False

    if gsa_ntax is None:
        terminate_at_full_tree = True
        # gsa_ntax = 1 + target_num_taxa
    elif target_num_extant_tips is None:
        raise ValueError("If 'gsa_ntax' is specified, 'num_extant_tips' must be specified")
    elif target_num_extinct_tips is not None:
        raise ValueError("If 'gsa_ntax' is specified, 'num_extinct_tips' cannot be specified")
    elif target_num_total_tips is not None:
        raise ValueError("If 'gsa_ntax' is specified, 'num_total_tips' cannot be specified")
    elif gsa_ntax < target_num_extant_tips:
        raise ValueError("'gsa_ntax' ({}) must be greater than 'num_extant_tips' ({})".format(gsa_ntax, target_num_extant_tips))

    # initialize tree
    if tree is not None:
        if taxon_namespace is not None:
            assert tree.taxon_namespace is taxon_namespace
        else:
            taxon_namespace = tree.taxon_namespace
        extant_tips = []
        extinct_tips = []
        for nd in tree:
            if not nd._child_nodes:
                if getattr(nd, extinct_attr_name, False):
                    extant_tips.append(nd)
                    if is_add_extinct_attr:
                        setattr(nd, extinct_attr_name, False)
                else:
                    extinct_tips.append(nd)
                    if is_add_extinct_attr:
                        setattr(nd, extinct_attr_name, True)
            elif is_add_extinct_attr:
                setattr(nd, extinct_attr_name, None)
    else:
        if taxon_namespace is None:
            taxon_namespace = dendropy.TaxonNamespace()
        tree = dendropy.Tree(taxon_namespace=taxon_namespace)
        tree.is_rooted = True
        tree.seed_node.edge.length = 0.0
        tree.seed_node.birth_rate = birth_rate
        tree.seed_node.death_rate = death_rate
        if is_add_extinct_attr:
            setattr(tree.seed_node, extinct_attr_name, False)
        extant_tips = [tree.seed_node]
        extinct_tips = []
    initial_extant_tip_set = list(extant_tips)
    initial_extinct_tip_set = list(extinct_tips)

    total_time = 0

    # for the GSA simulations targetted_time_slices is a list of tuple
    #   the first element in the tuple is the duration of the amount
    #   that the simulation spent at the (targetted) number of taxa
    #   and a list of edge information. The list of edge information includes
    #   a list of terminal edges in the tree and the length for that edge
    #   that marks the beginning of the time slice that corresponds to the
    #   targetted number of taxa.
    targetted_time_slices = []

    while True:
        if gsa_ntax is None:
            if target_num_extant_tips is not None and len(extant_tips) >= target_num_extant_tips:
                break
            if target_num_extinct_tips is not None and len(extinct_tips) >= target_num_extinct_tips:
                break
            if target_num_total_tips is not None and (len(extant_tips) + len(extinct_tips)) >= target_num_total_tips:
                break
            if max_time is not None and total_time >= max_time:
                break
        elif len(extant_tips) >= gsa_ntax:
            break

        # get vector of birth/death probabilities, and
        # associate with nodes/events
        event_rates = []
        event_nodes = []
        for nd in extant_tips:
            if not hasattr(nd, 'birth_rate'):
                nd.birth_rate = birth_rate
            if not hasattr(nd, 'death_rate'):
                nd.death_rate = death_rate
            event_rates.append(nd.birth_rate)
            event_nodes.append((nd, True)) # birth event = True
            event_rates.append(nd.death_rate)
            event_nodes.append((nd, False)) # birth event = False; i.e. death

        # get total probability of any birth/death
        rate_of_any_event = sum(event_rates)

        # waiting time based on above probability
        waiting_time = rng.expovariate(rate_of_any_event)

        if ( (gsa_ntax is not None)
                and (len(extant_tips) == target_num_extant_tips)
                ):
            edge_and_start_length = []
            for nd in extant_tips:
                e = nd.edge
                edge_and_start_length.append((e, e.length))
            targetted_time_slices.append((waiting_time, edge_and_start_length))
            if terminate_at_full_tree:
                break

        # add waiting time to nodes
        for nd in extant_tips:
            try:
                nd.edge.length += waiting_time
            except TypeError:
                nd.edge.length = waiting_time
        total_time += waiting_time

        # if event occurs within time constraints
        if max_time is None or total_time <= max_time:
            # normalize probability
            for i in range(len(event_rates)):
                event_rates[i] = event_rates[i]/rate_of_any_event
            # select node/event and process
            nd, birth_event = probability.weighted_choice(event_nodes, event_rates, rng=rng)
            extant_tips.remove(nd)
            if birth_event:
                if is_add_extinct_attr:
                    setattr(nd, extinct_attr_name, None)
                c1 = nd.new_child()
                c2 = nd.new_child()
                c1.edge.length = 0
                c2.edge.length = 0
                c1.birth_rate = nd.birth_rate + rng.gauss(0, birth_rate_sd)
                c1.death_rate = nd.death_rate + rng.gauss(0, death_rate_sd)
                c2.birth_rate = nd.birth_rate + rng.gauss(0, birth_rate_sd)
                c2.death_rate = nd.death_rate + rng.gauss(0, death_rate_sd)
                extant_tips.append(c1)
                extant_tips.append(c2)
            else:
                if len(extant_tips) > 0:
                    extinct_tips.append(nd)
                    if is_add_extinct_attr:
                        setattr(nd, extinct_attr_name, None)
                else:
                    # total extinction
                    if (gsa_ntax is not None):
                        if (len(targetted_time_slices) > 0):
                            break
                    if not repeat_until_success:
                        raise TreeSimTotalExtinctionException()
                    # We are going to basically restart the simulation because
                    # the tree has gone extinct (without reaching the specified
                    # ntax)
                    extant_tips = list(initial_extant_tip_set)
                    extinct_tips = list(initial_extinct_tip_set)
                    for nd in extant_tips:
                        if is_add_extinct_attr:
                            setattr(nd, extinct_attr_name, False)
                        nd.clear_child_nodes()
                    total_time = 0

    if gsa_ntax is not None:
        total_duration_at_target_n_tax = 0.0
        for i in targetted_time_slices:
            total_duration_at_target_n_tax += i[0]
        r = rng.random()*total_duration_at_target_n_tax
        selected_slice = None
        for n, i in enumerate(targetted_time_slices):
            r -= i[0]
            if r < 0.0:
                selected_slice = i
        assert(selected_slice is not None)
        edges_at_slice = selected_slice[1]
        last_waiting_time = selected_slice[0]

        for e, prev_length in edges_at_slice:
            daughter_nd = e.head_node
            for nd in daughter_nd.child_nodes():
                nd._parent_node = None
                try:
                    extinct_tips.remove(nd)
                except ValueError:
                    pass
                try:
                    extant_tips.remove(nd)
                except ValueError:
                    pass
                for desc in nd.preorder_iter():
                    try:
                        extant_tips.remove(desc)
                    except ValueError:
                        pass
            daughter_nd.clear_child_nodes()
            try:
                extinct_tips.remove(daughter_nd)
            except ValueError:
                pass
            extant_tips.append(daughter_nd)
            if is_add_extinct_attr:
                setattr(daughter_nd, extinct_attr_name, False)
            e.length = prev_length + last_waiting_time

    if not is_retain_extinct_tips:
        processed_nodes = set()
        for nd in list(extinct_tips):
            if nd in processed_nodes:
                continue
            processed_nodes.add(nd)
            try:
                extinct_tips.remove(nd)
            except ValueError:
                pass
            assert not nd._child_nodes
            while (nd.parent_node is not None) and (len(nd.parent_node._child_nodes) == 1):
                nd = nd.parent_node
                processed_nodes.add(nd)
            tree.prune_subtree(nd, suppress_unifurcations=False)
    tree.suppress_unifurcations()

    if is_assign_extant_taxa or is_assign_extinct_taxa:
        taxon_pool = [t for t in taxon_namespace]
        rng.shuffle(taxon_pool)
        taxon_pool_labels = set([t.label for t in taxon_pool])

        ### ONLY works if in GSA sub-section we remove ALL extant and
        ### extinct nodes beyond time slice: expensive
        ### Furthermore, main reason to use this approach is to have different
        ### label prefixes for extinct vs. extant lineages, but the second time
        ### this function is called with the same taxon namespace or any time
        ### this function is called with a populated taxon namespace, that
        ### aesthetic is lost.
        # node_pool_labels = ("T", "X")
        # for node_pool_idx, node_pool in enumerate((extant_tips, extinct_tips)):
        #     for node_idx, nd in enumerate(node_pool):
        #         if taxon_pool:
        #             taxon = taxon_pool.pop()
        #         else:
        #             taxon = taxon_namespace.require_taxon("{}{}".format(node_pool_labels[node_pool_idx], node_idx+1))
        #         nd.taxon = taxon
        #         assert not nd._child_nodes

        tlabel_counter = 0
        leaf_nodes = tree.leaf_nodes()
        rng.shuffle(leaf_nodes)
        for nd_idx, nd in enumerate(leaf_nodes):
            if not is_assign_extant_taxa and nd in extant_tips:
                continue
            if not is_assign_extant_taxa and nd in extinct_tips:
                continue
            if taxon_pool:
                taxon = taxon_pool.pop()
            else:
                while True:
                    tlabel_counter += 1
                    label = "{}{}".format("T", tlabel_counter)
                    if label not in taxon_pool_labels:
                        break
                taxon = taxon_namespace.require_taxon(label=label)
                taxon_pool_labels.add(label)
            nd.taxon = taxon
    return tree


#NICOLA: new version of birth-death tree simulation with linear time cost
#in number of tips instead of quadratic. To obtain this accelleration, however,
#I have removed variation in birth and death rates.
def fast_birth_death_tree(birth_rate, death_rate, **kwargs):
    #NICOLA: we are not allowing variation in birth and death rate so to have
    #increased efficiency.
    birth_rate_sd=0.0
    death_rate_sd=0.0

    """
    Returns a birth-death tree with birth rate specified by ``birth_rate``, and
    death rate specified by ``death_rate``, with edge lengths in continuous (real)
    units.

    Tree growth is controlled by one or more of the following arguments, of which
    at least one must be specified:

        - If ``num_extant_tips`` is given as a keyword argument, tree is grown until the
          number of EXTANT tips equals this number.
        - If ``num_extinct_tips`` is given as a keyword argument, tree is grown until the
          number of EXTINCT tips equals this number.
        - If ``num_total_tips`` is given as a keyword argument, tree is grown until the
          number of EXTANT plus EXTINCT tips equals this number.
        - If 'max_time' is given as a keyword argument, tree is grown for
          a maximum of ``max_time``.
        - If ``gsa_ntax`` is given then the tree will be simulated up to this number of
          EXTANT tips (or 0 tips), then a tree will be randomly selected from the
          intervals which corresond to times at which the tree had exactly ``num_extant_tips``
          leaves. This allows for simulations according to the "General
          Sampling Approach" of Hartmann et al. (2010). If this option is
          specified, then ``num_extant_tips`` MUST be specified and
          ``num_extinct_tips`` and ``num_total_tips`` CANNOT be specified.

    If more than one of the above is given, then tree growth will terminate when
    *any* one of the termination conditions are met.

    Parameters
    ----------

    birth_rate : float
        The birth rate.
    death_rate : float
        The death rate.
    birth_rate_sd : float
        The standard deviation of the normally-distributed mutation added to
        the birth rate as it is inherited by daughter nodes; if 0, birth rate
        does not evolve on the tree.
    death_rate_sd : float
        The standard deviation of the normally-distributed mutation added to
        the death rate as it is inherited by daughter nodes; if 0, death rate
        does not evolve on the tree.

    Keyword Arguments
    -----------------

    num_extant_tips: int
        If specified, branching process is terminated when number of EXTANT
        tips equals this number.
    num_extinct_tips: int
        If specified, branching process is terminated when number of EXTINCT
        tips equals this number.
    num_total_tips: int
        If specified, branching process is terminated when number of EXTINCT
        plus EXTANT tips equals this number.
    max_time: float
        If specified, branching process is terminated when time reaches or
        exceeds this value.
    gsa_ntax: int
        The General Sampling Approach threshold for number of taxa. See above
        for details.
    tree : Tree instance
        If given, then this tree will be used; otherwise a new one will be created.
    taxon_namespace : TaxonNamespace instance
        If given, then this will be assigned to the new tree, and, in addition,
        taxa assigned to tips will be sourced from or otherwise created with
        reference to this.
    is_assign_extant_taxa : bool [default: True]
        If False, then taxa will not be assigned to extant tips. If True
        (default), then taxa will be assigned to extant tips. Taxa will be
        assigned from the specified ``taxon_namespace`` or
        ``tree.taxon_namespace``. If the number of taxa required exceeds the
        number of taxa existing in the taxon namespace, new |Taxon| objects
        will be created as needed and added to the taxon namespace.
    is_assign_extinct_taxa : bool [default: True]
        If False, then taxa will not be assigned to extant tips. If True
        (default), then taxa will be assigned to extant tips. Taxa will be
        assigned from the specified ``taxon_namespace`` or
        ``tree.taxon_namespace``. If the number of taxa required exceeds the
        number of taxa existing in the taxon namespace, new |Taxon| objects
        will be created as needed and added to the taxon namespace. Note that
        this option only makes sense if extinct tips are retained (specified via
        'is_retain_extinct_tips' option), and will otherwise be ignored.
    is_add_extinct_attr: bool [default: True]
        If True (default), add an boolean attribute indicating whether or not a
        node is an extinct tip or not. False will skip this. Name of attribute
        is set by 'extinct_attr_name' argument, defaulting to 'is_extinct'.
        Note that this option only makes sense if extinct tips are retained
        (specified via 'is_retain_extinct_tips' option), and will otherwise be
        ignored.
    extinct_attr_name: str [default: 'is_extinct']
        Name of attribute to add to nodes indicating whether or not tip is extinct.
        Note that this option only makes sense if extinct tips are retained
        (specified via 'is_retain_extinct_tips' option), and will otherwise be
        ignored.
    is_retain_extinct_tips : bool [default: False]
        If True, extinct tips will be retained on tree. Defaults to False:
        extinct lineages removed from tree.
    repeat_until_success: bool [default: True]
        Under some conditions, it is possible for all lineages on a tree to go
        extinct. In this case, if this argument is given as |True| (the
        default), then a new branching process is initiated. If |False|
        (default), then a TreeSimTotalExtinctionException is raised.
    rng: random.Random() or equivalent instance
        A Random() object or equivalent can be passed using the ``rng`` keyword;
        otherwise GLOBAL_RNG is used.

    References
    ----------

    Hartmann, Wong, and Stadler "Sampling Trees from Evolutionary Models" Systematic Biology. 2010. 59(4). 465-476

    """
    if "assign_taxa" in kwargs:
        deprecate.dendropy_deprecation_warning(
                message="Deprecated: 'assign_taxa' will no longer be supported as an argument to this function. Use 'is_assign_extant_taxa' and/or 'is_assign_extinct_taxa' instead",
                stacklevel=3)
        a = kwargs.pop("assign_taxa")
        kwargs["is_assign_extant_taxa"] = a
        kwargs["is_assign_extant_taxa"] = a
    if "ntax" in kwargs:
        deprecate.dendropy_deprecation_warning(
                message="Deprecated: 'ntax' is no longer supported as an argument to this function. Use one or more of the following instead: 'num_extant_tips', 'num_extinct_tips', 'num_total_tips', or 'max_time'",
                stacklevel=3)
        kwargs["num_extant_tips"] = kwargs.pop("ntax")
    if (("num_extant_tips" not in kwargs)
            and ("num_extinct_tips" not in kwargs)
            and ("num_total_tips" not in kwargs)
            and ("max_time" not in kwargs) ):
        if "taxon_namespace" in kwargs:
            ### cannot support legacy approach, b/c ``taxon_namespace`` may grow during function, leading to unpredictable behavior
            # deprecate.dendropy_deprecation_warning(
            #         preamble="Deprecated: The 'taxon_namespace' argument can no longer be used to specify a termination condition as a side-effect. Use one or more of the following instead with the length of the taxon namespace instance as a value: 'num_extant_tips', 'num_extinct_tips', or 'num_total_tips'",
            #         old_construct="tree = birth_death_tree(\n    ...\n    taxon_namespace=taxon_namespace,\n    ...\n)",
            #         new_construct="tree = birth_death_tree(\n    ...\n    taxon_namespace=taxon_namespace,\n    num_extant_tips=len(taxon_namespace),\n    ...\n)")
            # kwargs["num_extant_tips"] = len(kwargs["taxon_namespace"])
            raise ValueError("The 'taxon_namespace' argument can no longer be used to specify a termination condition as a side-effect."
                             "Use one or more of the following instead with the length of the taxon namespace instance as a value: "
                             "'num_extant_tips', 'num_extinct_tips', or 'num_total_tips'.\n"
                             "That is, instead of:\n\n"
                             "    tree = birth_death_tree(\n        ...\n        taxon_namespace=taxon_namespace,\n        ...\n    )\n\n"
                             "Use:\n\n"
                             "    ntax = len(taxon_namespace)\n    tree = birth_death_tree(\n        ...\n        taxon_namespace=taxon_namespace,\n        num_extant_tips=ntax,\n        ...\n    )\n"
                             "\nOr (recommended):\n\n"
                             "    tree = birth_death_tree(\n        ...\n        taxon_namespace=taxon_namespace,\n        num_extant_tips=100,\n        ...\n    )\n"
                             "\nNote that the taxon namespace instance size may grow during any particular call of the function depending on taxon assignment/creation settings, so"
                             " for stable and predictable behavor it is important to take a snapshot of the desired taxon namespace size before any call of the function, or, better yet"
                             " simply pass in a constant value."
                             )
        else:
            raise ValueError("One or more of the following must be specified: 'num_extant_tips', 'num_extinct_tips', or 'max_time'")
    target_num_extant_tips = kwargs.pop("num_extant_tips", None)
    target_num_extinct_tips = kwargs.pop("num_extinct_tips", None)
    target_num_total_tips = kwargs.pop("num_total_tips", None)
    max_time = kwargs.pop('max_time', None)
    gsa_ntax = kwargs.pop('gsa_ntax', None)
    is_add_extinct_attr = kwargs.pop('is_add_extinct_attr', True)
    extinct_attr_name = kwargs.pop('extinct_attr_name', 'is_extinct')
    is_retain_extinct_tips = kwargs.pop('is_retain_extinct_tips', False)
    is_assign_extant_taxa = kwargs.pop('is_assign_extant_taxa', True)
    is_assign_extinct_taxa = kwargs.pop('is_assign_extinct_taxa', True)
    repeat_until_success = kwargs.pop('repeat_until_success', True)

    tree = kwargs.pop("tree", None)
    taxon_namespace = kwargs.pop("taxon_namespace", None)

    rng = kwargs.pop('rng', GLOBAL_RNG)

    ignore_unrecognized_keyword_arguments = kwargs.pop('ignore_unrecognized_keyword_arguments', False)
    if kwargs and not ignore_unrecognized_keyword_arguments:
        raise ValueError("Unsupported keyword arguments: {}".format(kwargs.keys()))

    terminate_at_full_tree = False

    if gsa_ntax is None:
        terminate_at_full_tree = True
        # gsa_ntax = 1 + target_num_taxa
    elif target_num_extant_tips is None:
        raise ValueError("If 'gsa_ntax' is specified, 'num_extant_tips' must be specified")
    elif target_num_extinct_tips is not None:
        raise ValueError("If 'gsa_ntax' is specified, 'num_extinct_tups' cannot be specified")
    elif target_num_total_tips is not None:
        raise ValueError("If 'gsa_ntax' is specified, 'num_total_tips' cannot be specified")
    elif gsa_ntax < target_num_extant_tips:
        raise ValueError("'gsa_ntax' ({}) must be greater than 'num_extant_tips' ({})".format(gsa_ntax, target_num_extant_tips))

    # initialize tree
    if tree is not None:
        if taxon_namespace is not None:
            assert tree.taxon_namespace is taxon_namespace
        else:
            taxon_namespace = tree.taxon_namespace
        extant_tips = []
        extinct_tips = []
        for nd in tree:
            if not nd._child_nodes:
                if getattr(nd, extinct_attr_name, False):
                    extant_tips.append(nd)
                    #NICOLA: the reason for this will become clear later
                    nd.edge.length=-nd.edge.length
                    if is_add_extinct_attr:
                        setattr(nd, extinct_attr_name, False)
                else:
                    extinct_tips.append(nd)
                    if is_add_extinct_attr:
                        setattr(nd, extinct_attr_name, True)
            elif is_add_extinct_attr:
                setattr(nd, extinct_attr_name, None)
    else:
        if taxon_namespace is None:
            taxon_namespace = dendropy.TaxonNamespace()
        tree = dendropy.Tree(taxon_namespace=taxon_namespace)
        tree.is_rooted = True
        tree.seed_node.edge.length = 0.0
        tree.seed_node.birth_rate = birth_rate
        tree.seed_node.death_rate = death_rate
        if is_add_extinct_attr:
            setattr(tree.seed_node, extinct_attr_name, False)
        extant_tips = [tree.seed_node]
        extinct_tips = []
    initial_extant_tip_set = list(extant_tips)
    initial_extinct_tip_set = list(extinct_tips)
    #NICOLA: in case we need a restart, let's record the initial branch lengths.
    initial_lengths=[]
    for nd in initial_extant_tip_set:
        initial_lengths.append(nd.edge.length)

    total_time = 0

    # for the GSA simulations targetted_time_slices is a list of tuple
    #   the first element in the tuple is the duration of the amount
    #   that the simulation spent at the (targetted) number of taxa
    #   and a list of edge information. The list of edge information includes
    #   a list of terminal edges in the tree and the length for that edge
    #   that marks the beginning of the time slice that corresponds to the
    #   targetted number of taxa.
    targetted_time_slices = []

    while True:
        if gsa_ntax is None:
            if target_num_extant_tips is not None and len(extant_tips) >= target_num_extant_tips:
                #NICOLA: here and in the other cases, update the branch length so to
                #represent the true branch length (explained better below).
                for nd in extant_tips:
                        nd.edge.length=total_time-nd.edge.length
                break
            if target_num_extinct_tips is not None and len(extinct_tips) >= target_num_extinct_tips:
                for nd in extant_tips:
                        nd.edge.length=total_time-nd.edge.length
                break
            if target_num_total_tips is not None and (len(extant_tips) + len(extinct_tips)) >= target_num_total_tips:
                for nd in extant_tips:
                        nd.edge.length=total_time-nd.edge.length
                break
            if max_time is not None and total_time >= max_time:
                for nd in extant_tips:
                        nd.edge.length=total_time-nd.edge.length
                break
        elif len(extant_tips) >= gsa_ntax:
                for nd in extant_tips:
                        nd.edge.length=total_time-nd.edge.length
                break


        # get vector of birth/death probabilities, and
        # associate with nodes/events
        #NICOLA: replaced this
   #      event_rates = []
#         event_nodes = []
#         for nd in extant_tips:
#             if not hasattr(nd, 'birth_rate'):
#                 nd.birth_rate = birth_rate
#             if not hasattr(nd, 'death_rate'):
#                 nd.death_rate = death_rate
#             event_rates.append(nd.birth_rate)
#             event_nodes.append((nd, True)) # birth event = True
#             event_rates.append(nd.death_rate)
#             event_nodes.append((nd, False)) # birth event = False; i.e. death


        # get total probability of any birth/death
        #rate_of_any_event = sum(event_rates)
        #NICOLA: instead, assume that each node has the same birth and death rate, and sample
        #from the total rate and then sample one node at random from extant ones.
        #this has constant cost instead of linear in the number of extant taxa.
        rate_of_any_event=len(extant_tips)*(birth_rate+death_rate)

        # waiting time based on above probability
        waiting_time = rng.expovariate(rate_of_any_event)

        if ( (gsa_ntax is not None)
                and (len(extant_tips) == target_num_extant_tips)
                ):
            edge_and_start_length = []
            for nd in extant_tips:
                e = nd.edge
                #NICOLA: modified accordingly
                edge_and_start_length.append((e, total_time-e.length))
            targetted_time_slices.append((waiting_time, edge_and_start_length))
            if terminate_at_full_tree:
                #NICOLA: terminal update as before
                for nd in extant_tips:
                        nd.edge.length=total_time-nd.edge.length
                break

        # add waiting time to nodes
        #Nicola: this is the part that makes it quadratic in ntax.
        #instead, we use a trick of initializing nd.edge.length to the time of node creation,
        # and then we update it to the correct branch length when the node is closed or simulations are terminated.
        #for nd in extant_tips:
        #    try:
        #        nd.edge.length += waiting_time
        #    except TypeError:
        #        nd.edge.length = waiting_time
        total_time += waiting_time

        # if event occurs within time constraints
        if max_time is None or total_time <= max_time:
            # normalize probability
            #for i in range(len(event_rates)):
            #    event_rates[i] = event_rates[i]/rate_of_any_event
            # select node/event and process
            #nd, birth_event = probability.weighted_choice(event_nodes, event_rates, rng=rng)
            #NICOLA: instead of sampling from a categorical distribution (requiring linear time in ntax)
            #sample a random integer to represent the chosen taxon, and sample if birth or death event.
            taxI=rng.randint(0, len(extant_tips)-1)
            if rng.random()< birth_rate/(birth_rate+death_rate):
                birth_event=True
            else:
                birth_event=False
            nd=extant_tips[taxI]
            #extant_tips.remove(nd)
            if birth_event:
                if is_add_extinct_attr:
                    setattr(nd, extinct_attr_name, None)
                c1 = nd.new_child()
                c2 = nd.new_child()
                extant_tips[taxI]=c1
                #NICOLA: we initialize branch lengths to total time;
                #this may not make sense at first, but when we are done with a node, we
                #update the branch length to the new current time minus the time of node creation
                #which gives us the branch length. The advantage is that this takes linear time
                #instead of quadratic.
                c1.edge.length = total_time
                c2.edge.length = total_time
                nd.edge.length = total_time-nd.edge.length
                #c1.edge.length = 0
                #c2.edge.length = 0
                #NICOLA: here, for speed, we don't allow variation in birth rates.
                #It could be possible to allow it, but it would take ntax*log(ntax) time,
                #which wouldn't be bad, but it would also require a more complicated approach.
                #c1.birth_rate = nd.birth_rate + rng.gauss(0, birth_rate_sd)
                #c1.death_rate = nd.death_rate + rng.gauss(0, death_rate_sd)
                #c2.birth_rate = nd.birth_rate + rng.gauss(0, birth_rate_sd)
                #c2.death_rate = nd.death_rate + rng.gauss(0, death_rate_sd)
                c1.birth_rate = nd.birth_rate
                c1.death_rate = nd.death_rate
                c2.birth_rate = nd.birth_rate
                c2.death_rate = nd.death_rate
                #extant_tips.append(c1)
                extant_tips.append(c2)
            else:
                del extant_tips[taxI]
                if len(extant_tips) > 0:
                    extinct_tips.append(nd)
                    if is_add_extinct_attr:
                        setattr(nd, extinct_attr_name, None)
                else:
                    # total extinction
                    if (gsa_ntax is not None):
                        if (len(targetted_time_slices) > 0):
                            break
                    if not repeat_until_success:
                        raise TreeSimTotalExtinctionException()
                    # We are going to basically restart the simulation because
                    # the tree has gone extinct (without reaching the specified
                    # ntax)
                    extant_tips = list(initial_extant_tip_set)
                    extinct_tips = list(initial_extinct_tip_set)
                    ndIndex=0
                    for nd in extant_tips:
                        if is_add_extinct_attr:
                            setattr(nd, extinct_attr_name, False)
                        nd.clear_child_nodes()
                        #NICOLA: restore branch length to original value
                        nd.edge.length=initial_lengths[ndIndex]
                        ndIndex+=1
                    total_time = 0


    if gsa_ntax is not None:
        total_duration_at_target_n_tax = 0.0
        for i in targetted_time_slices:
            total_duration_at_target_n_tax += i[0]
        r = rng.random()*total_duration_at_target_n_tax
        selected_slice = None
        for n, i in enumerate(targetted_time_slices):
            r -= i[0]
            if r < 0.0:
                selected_slice = i
        assert(selected_slice is not None)
        edges_at_slice = selected_slice[1]
        last_waiting_time = selected_slice[0]

        for e, prev_length in edges_at_slice:
            daughter_nd = e.head_node
            for nd in daughter_nd.child_nodes():
                nd._parent_node = None
                try:
                    extinct_tips.remove(nd)
                except ValueError:
                    pass
                try:
                    extant_tips.remove(nd)
                except ValueError:
                    pass
                for desc in nd.preorder_iter():
                    try:
                        extant_tips.remove(desc)
                    except ValueError:
                        pass
            daughter_nd.clear_child_nodes()
            try:
                extinct_tips.remove(daughter_nd)
            except ValueError:
                pass
            extant_tips.append(daughter_nd)
            if is_add_extinct_attr:
                setattr(daughter_nd, extinct_attr_name, False)
            e.length = prev_length + last_waiting_time

    if not is_retain_extinct_tips:
        processed_nodes = set()
        for nd in list(extinct_tips):
            if nd in processed_nodes:
                continue
            processed_nodes.add(nd)
            try:
                extinct_tips.remove(nd)
            except ValueError:
                pass
            assert not nd._child_nodes
            while (nd.parent_node is not None) and (len(nd.parent_node._child_nodes) == 1):
                nd = nd.parent_node
                processed_nodes.add(nd)
            tree.prune_subtree(nd, suppress_unifurcations=False)
    tree.suppress_unifurcations()

    if is_assign_extant_taxa or is_assign_extinct_taxa:
        taxon_pool = [t for t in taxon_namespace]
        rng.shuffle(taxon_pool)
        taxon_pool_labels = set([t.label for t in taxon_pool])

        ### ONLY works if in GSA sub-section we remove ALL extant and
        ### extinct nodes beyond time slice: expensive
        ### Furthermore, main reason to use this approach is to have different
        ### label prefixes for extinct vs. extant lineages, but the second time
        ### this function is called with the same taxon namespace or any time
        ### this function is called with a populated taxon namespace, that
        ### aesthetic is lost.
        # node_pool_labels = ("T", "X")
        # for node_pool_idx, node_pool in enumerate((extant_tips, extinct_tips)):
        #     for node_idx, nd in enumerate(node_pool):
        #         if taxon_pool:
        #             taxon = taxon_pool.pop()
        #         else:
        #             taxon = taxon_namespace.require_taxon("{}{}".format(node_pool_labels[node_pool_idx], node_idx+1))
        #         nd.taxon = taxon
        #         assert not nd._child_nodes

        tlabel_counter = 0
        leaf_nodes = tree.leaf_nodes()
        rng.shuffle(leaf_nodes)
        for nd_idx, nd in enumerate(leaf_nodes):
            if not is_assign_extant_taxa and nd in extant_tips:
                continue
            if not is_assign_extant_taxa and nd in extinct_tips:
                continue
            if taxon_pool:
                taxon = taxon_pool.pop()
            else:
                while True:
                    tlabel_counter += 1
                    label = "{}{}".format("T", tlabel_counter)
                    if label not in taxon_pool_labels:
                        break
                taxon = taxon_namespace.require_taxon(label=label)
                taxon_pool_labels.add(label)
            nd.taxon = taxon
    return tree



def discrete_birth_death_tree(birth_rate, death_rate, birth_rate_sd=0.0, death_rate_sd=0.0, **kwargs):
    """
    Returns a birth-death tree with birth rate specified by ``birth_rate``, and
    death rate specified by ``death_rate``, with edge lengths in discrete (integer)
    units.

    ``birth_rate_sd`` is the standard deviation of the normally-distributed mutation
    added to the birth rate as it is inherited by daughter nodes; if 0, birth
    rate does not evolve on the tree.

    ``death_rate_sd`` is the standard deviation of the normally-distributed mutation
    added to the death rate as it is inherited by daughter nodes; if 0, death
    rate does not evolve on the tree.

    Tree growth is controlled by one or more of the following arguments, of which
    at least one must be specified:

        - If ``ntax`` is given as a keyword argument, tree is grown until the number of
          tips == ntax.
        - If ``taxon_namespace`` is given as a keyword argument, tree is grown until the
          number of tips == len(taxon_namespace), and the taxa are assigned randomly to the
          tips.
        - If 'max_time' is given as a keyword argument, tree is grown for ``max_time``
          number of generations.

    If more than one of the above is given, then tree growth will terminate when
    *any* of the termination conditions (i.e., number of tips == ``ntax``, or number
    of tips == len(taxon_namespace) or number of generations = ``max_time``) are met.

    Also accepts a Tree object (with valid branch lengths) as an argument passed
    using the keyword ``tree``: if given, then this tree will be used; otherwise
    a new one will be created.

    If ``assign_taxa`` is False, then taxa will *not* be assigned to the tips;
    otherwise (default), taxa will be assigned. If ``taxon_namespace`` is given
    (``tree.taxon_namespace``, if ``tree`` is given), and the final number of tips on the
    tree after the termination condition is reached is less then the number of
    taxa in ``taxon_namespace`` (as will be the case, for example, when
    ``ntax`` < len(``taxon_namespace``)), then a random subset of taxa in ``taxon_namespace`` will
    be assigned to the tips of tree. If the number of tips is more than the number
    of taxa in the ``taxon_namespace``, new Taxon objects will be created and added
    to the ``taxon_namespace`` if the keyword argument ``create_required_taxa`` is not given as
    False.

    Under some conditions, it is possible for all lineages on a tree to go extinct.
    In this case, if the keyword argument ``repeat_until_success`` is |True|, then a new
    branching process is initiated.
    If |False| (default), then a TreeSimTotalExtinctionException is raised.

    A Random() object or equivalent can be passed using the ``rng`` keyword;
    otherwise GLOBAL_RNG is used.
    """
    if 'ntax' not in kwargs \
        and 'taxon_namespace' not in kwargs \
        and 'max_time' not in kwargs:
            raise ValueError("At least one of the following must be specified: 'ntax', 'taxon_namespace', or 'max_time'")
    target_num_taxa = None
    taxon_namespace = None
    target_num_gens = kwargs.get('max_time', None)
    if 'taxon_namespace' in kwargs:
        taxon_namespace = kwargs.get('taxon_namespace')
        target_num_taxa = kwargs.get('ntax', len(taxon_namespace))
    elif 'ntax' in kwargs:
        target_num_taxa = kwargs['ntax']
    if taxon_namespace is None:
        taxon_namespace = dendropy.TaxonNamespace()
    repeat_until_success = kwargs.get('repeat_until_success', False)
    rng = kwargs.get('rng', GLOBAL_RNG)

    # grow tree
    if "tree" in kwargs:
        tree = kwargs['tree']
        if "taxon_namespace" in kwargs and kwargs['taxon_namespace'] is not tree.taxon_namespace:
            raise ValueError("Cannot specify both ``tree`` and ``taxon_namespace``")
    else:
        tree = dendropy.Tree(taxon_namespace=taxon_namespace)
        tree.is_rooted = True
        tree.seed_node.edge.length = 0
        tree.seed_node.birth_rate = birth_rate
        tree.seed_node.death_rate = death_rate
    leaf_nodes = tree.leaf_nodes()
    num_gens = 0
    while (target_num_taxa is None or len(leaf_nodes) < target_num_taxa) \
            and (target_num_gens is None or num_gens < target_num_gens):
        for nd in leaf_nodes:
            if not hasattr(nd, 'birth_rate'):
                nd.birth_rate = birth_rate
            if not hasattr(nd, 'death_rate'):
                nd.death_rate = death_rate
            try:
                nd.edge.length += 1
            except TypeError:
                nd.edge.length = 1
            u = rng.uniform(0, 1)
            if u < nd.birth_rate:
                c1 = nd.new_child()
                c2 = nd.new_child()
                c1.edge.length = 0
                c2.edge.length = 0
                c1.birth_rate = nd.birth_rate + rng.gauss(0, birth_rate_sd)
                c1.death_rate = nd.death_rate + rng.gauss(0, death_rate_sd)
                c2.birth_rate = nd.birth_rate + rng.gauss(0, birth_rate_sd)
                c2.death_rate = nd.death_rate + rng.gauss(0, death_rate_sd)
            elif u > nd.birth_rate and u < (nd.birth_rate + nd.death_rate):
                if nd is not tree.seed_node:
                    tree.prune_subtree(nd)
                elif not repeat_until_success:
                    # all lineages are extinct: raise exception
                    raise TreeSimTotalExtinctionException()
                else:
                    # all lineages are extinct: repeat
                    num_gens = 0

        num_gens += 1
        leaf_nodes = tree.leaf_nodes()

    # If termination condition specified by ntax or taxon_namespace, then the last
    # split will have a daughter edges of length == 0;
    # so we continue growing the edges until the next birth/death event *or*
    # the max number of generations condition is given and met
    gens_to_add = 0
    while (target_num_gens is None or num_gens < target_num_gens):
        u = rng.uniform(0, 1)
        if u < (birth_rate + death_rate):
            break
        gens_to_add += 1
    for nd in tree.leaf_nodes():
        nd.edge.length += gens_to_add

    if kwargs.get("assign_taxa", True):
        tree.randomly_assign_taxa(create_required_taxa=True, rng=rng)

    # return
    return tree

def uniform_pure_birth_tree(taxon_namespace, birth_rate=1.0, rng=None):
    "Generates a uniform-rate pure-birth process tree. "
    if rng is None:
        rng = GLOBAL_RNG # use the global rng by default
    tree = dendropy.Tree(taxon_namespace=taxon_namespace)
    tree.seed_node.edge.length = 0.0
    leaf_nodes = tree.leaf_nodes()
    while len(leaf_nodes) < len(taxon_namespace):
        waiting_time = rng.expovariate(len(leaf_nodes)/birth_rate)
        for nd in leaf_nodes:
            nd.edge.length += waiting_time
        parent_node = rng.choice(leaf_nodes)
        c1 = parent_node.new_child()
        c2 = parent_node.new_child()
        c1.edge.length = 0.0
        c2.edge.length = 0.0
        leaf_nodes = tree.leaf_nodes()
    leaf_nodes = tree.leaf_nodes()
    waiting_time = rng.expovariate(len(leaf_nodes)/birth_rate)
    for nd in leaf_nodes:
        nd.edge.length += waiting_time
    for idx, leaf in enumerate(leaf_nodes):
        leaf.taxon = taxon_namespace[idx]
    tree.is_rooted = True
    return tree

def fit_pure_birth_model(**kwargs):
    r"""
    Calculates the maximum-likelihood estimate of the birth rate of a set of *internal* node ages under a Yule (pure-birth) model.

    Requires either a |Tree| object or an interable of *internal* node ages to be passed in via keyword arguments ``tree`` or ``internal_node_ages``, respectively. The former is more convenient when doing one-off calculations, while the latter is more efficient if the list of internal node ages needs to be used in other places and you already have it calculated and want to avoid re-calculating it here.

    Parameters
    ----------
    \*\*kwargs : keyword arguments, mandatory

        Exactly *one* of the following *must* be specified:

            tree : a |Tree| object.
                A |Tree| object. The tree needs to be ultrametric for the internal node ages (time from each internal node to the tips) to make sense. The precision by which the ultrametricity is checked can be specified using the ``ultrametricity_precision`` keyword argument (see below). If ``tree`` is given, then ``internal_node_ages`` cannot be given, and vice versa. If ``tree`` is not given, then ``internal_node_ages`` must be given.
            internal_node_ages : iterable (of numerical values)
                Iterable of node ages of the internal nodes of a tree, i.e., the list of sum of the edge lengths between each internal node and the tips of the tree. If ``internal_node_ages`` is given, then ``tree`` cannot be given, and vice versa. If ``internal_node_ages`` is not given, then ``tree`` must be given.

        The following are optional, and are only used if internal node ages are
        specified (i.e., 'internal_node_ages' are passed in):

            is_node_ages_presorted : bool
                By default, the vector of node ages are sorted. If this
                argument is specified as ``True``, then this sorting will be
                skipped, in which case it is the client code's responsibility
                to make sure that the node ages are given in REVERSE order
                (i.e., oldest nodes -- nodes closer to the root -- given
                first).
    Returns
    -------
    m : dictionary

    A dictionary with keys being parameter names and values being
    estimates:

        "birth_rate"
            The birth rate.
        "log_likelihood"
            The log-likelihood of the model and given birth rate.

    Examples
    --------

    Birth rates can be estimated by passing in trees directly::

        for idx, tree in enumerate(trees):
            m = birthdeath.fit_pure_birth_model(tree=tree)
            print("Tree {}: birth rate = {} (logL = {})".format(
                idx+1, m["birth_rate"], m["log_likelihood"]))

    Or by pre-calculating and passing in a list of node ages::

        for idx, tree in enumerate(trees):
            m = birthdeath.fit_pure_birth_model(
                    internal_node_ages=tree.internal_node_ages())
            print("Tree {}: birth rate = {} (logL = {})".format(
                idx+1, m["birth_rate"], m["log_likelihood"]))


    Notes
    -----
    Adapted from the laser package for R:

        -   Dan Rabosky and Klaus Schliep (2013). laser: Likelihood Analysis of
            Speciation/Extinction Rates from Phylogenies. R package version
            2.4-1. http://CRAN.R-project.org/package=laser

    """
    tree = kwargs.get("tree", None)
    if tree is not None:
        internal_node_ages = tree.internal_node_ages(ultrametricity_precision=kwargs.get("ultrametricity_precision", 1e-6))
    else:
        try:
            internal_node_ages = kwargs["internal_node_ages"]
        except KeyError:
            raise TypeError("Need to specify 'tree' or 'internal_node_ages'")
    x = internal_node_ages
    if tree is not None or not kwargs.get("is_node_ages_presorted", False):
        x = sorted(internal_node_ages, reverse=True)
    st1 = x[0]
    st2 = 0
    nvec = range(2, len(x)+2)
    nv = [i for i in x if (i < st1) and (i >= st2)]
    lo = max(nvec[idx] for idx, i in enumerate(x) if i >= st1)
    up = max(nvec[idx] for idx, i in enumerate(x) if i >= st2)
    if st1 <= x[0]:
        nv.insert(0, st1)
        nv = [i - st2 for i in nv]
    else:
        nv = [i - st2 for i in nv]
    t1 = (up-lo)
    t2 = (lo*nv[0])
    t3 = sum( nv[1:(up-lo+1)] )
    smax = t1/(t2 + t3)

    try:
        s1 = sum(map(math.log, range(lo,up)))
        s2 = (up-lo) * math.log(smax)
        s3 = lo - up
        lh = s1 + s2 + s3
    except ValueError:
        if kwargs.get("ignore_likelihood_calculation_failure", False):
            raise ValueError("Likelihood estimation failure")
        else:
            lh = float("-inf")

    result = {
        "birth_rate" : smax,
        "log_likelihood" : lh,
    }
    return result

def fit_pure_birth_model_to_tree(tree, ultrametricity_precision=constants.DEFAULT_ULTRAMETRICITY_PRECISION):
    """
    Calculates the maximum-likelihood estimate of the birth rate a tree under a
    Yule (pure-birth) model.

    Parameters
    ----------
    tree : |Tree| object
        A tree to be fitted.

    Returns
    -------
    m : dictionary

    A dictionary with keys being parameter names and values being
    estimates:

        -   "birth_rate"
            The birth rate.
        -   "log_likelihood"
            The log-likelihood of the model and given birth rate.

    Examples
    --------

    ::

        import dendropy
        from dendropy.model import birthdeath
        trees = dendropy.TreeList.get_from_path(
                "pythonidae.nex", "nexus")
        for idx, tree in enumerate(trees):
            m = birthdeath.fit_pure_birth_model_to_tree(tree)
            print("Tree {}: birth rate = {} (logL = {})".format(
                idx+1, m["birth_rate"], m["log_likelihood"]))

    """
    return fit_pure_birth_model(tree=tree, ultrametricity_precision=ultrametricity_precision)


def birth_death_likelihood(**kwargs):
    r"""
    Calculates the log-likelihood of a tree (or a set of internal nodes) under
    a birth death model.

    Requires either a |Tree| object or an interable of *internal* node
    ages to be passed in via keyword arguments ``tree`` or ``internal_node_ages``,
    respectively. The former is more convenient when doing one-off
    calculations, while the latter is more efficient if the list of internal
    node ages needs to be used in other places and you already have it
    calculated and want to avoid re-calculating it here.

    Parameters
    ----------
    \*\*kwargs : keyword arguments, mandatory

        Exactly *one* of the following *must* be specified:

            tree : a |Tree| object.
                A |Tree| object. The tree needs to be ultrametric for the
                internal node ages (time from each internal node to the tips)
                to make sense. The precision by which the ultrametricity is
                checked can be specified using the ``ultrametricity_precision`` keyword
                argument (see below). If ``tree`` is given, then
                ``internal_node_ages`` cannot be given, and vice versa. If ``tree``
                is not given, then ``internal_node_ages`` must be given.
            internal_node_ages : iterable (of numerical values)
                Iterable of node ages of the internal nodes of a tree, i.e., the
                list of sum of the edge lengths between each internal node and
                the tips of the tree. If ``internal_node_ages`` is given, then
                ``tree`` cannot be given, and vice versa. If ``internal_node_ages``
                is not given, then ``tree`` must be given.

        The following keyword parameters are mandatory:

            birth_rate : float
                The birth rate.
            death_rate : float
                The death rate.

        The following keyword parameters are optional:

            sampling_probability
                The probability for a species to be included in the sample. Defaults to 1.0 (all species sampled).
            sampling_strategy
                The strategy how samples were obtained. Options are: uniform|diversified|age.
            is_mrca_included
                Does the process start with the most recent common ancestor?
            condition_on : string
                Do we condition the process on: "time", "survival", or "taxa"?

        The following are optional, and are only used if internal node ages
        need to be calculated (i.e., 'tree' is passed in).

            ultrametricity_precision : float
                When calculating the node ages, an error will be raised if the tree in
                o ultrametric. This error may be due to floating-point or numerical
                imprecision. You can set the precision of the ultrametricity validation
                by setting the ``ultrametricity_precision`` parameter. E.g., use
                ``ultrametricity_precision=0.01`` for a more relaxed precision,
                down to 2 decimal places. Use ``ultrametricity_precision=False``
                to disable checking of ultrametricity precision.

            ignore_likelihood_calculation_failure: bool (default: False)
                In some cases (typically, abnormal trees, e.g., 1-tip), the
                likelihood estimation will fail. In this case a ValueError will
                be raised. If ``ignore_likelihood_calculation_failure`` is
                |True|, then the function call will still succeed, with the
                likelihood set to -``inf``.

        The following are optional, and are only used if internal node ages are
        specified (i.e., 'internal_node_ages' are passed in):

            is_node_ages_presorted : bool
                By default, the vector of node ages are sorted. If this
                argument is specified as ``True``, then this sorting will be
                skipped, in which case it is the client code's responsibility
                to make sure that the node ages are given in REVERSE order
                (i.e., oldest nodes -- nodes closer to the root -- given
                first).

    Notes
    -----
    Lifted directly from the (fantastic!) TESS package for R:

        H{"o}hna S. 2013. Fast simulation of reconstructed phylogenies under
        global time-dependent birth--death processes. Bioinformatics, 29(11)
        1367-1374.

    Returns
    -------
    lnl : float

    The log-likehood of the tree under the birth-death model.

    """
    tree = kwargs.get("tree", None)
    if tree is not None:
        internal_node_ages = tree.internal_node_ages(ultrametricity_precision=kwargs.get("ultrametricity_precision", 1e-6))
    else:
        try:
            internal_node_ages = kwargs["internal_node_ages"]
        except KeyError:
            raise TypeError("Need to specify 'tree' or 'internal_node_ages'")
    if tree is not None or not kwargs.get("is_node_ages_presorted", False):
        internal_node_ages = sorted(internal_node_ages, reverse=True)

    # check for sensible parameter values
    # if ( lambda <= 0 || mu < 0 || sampling_probability <= 0 || sampling_probability > 1.0) {
    #     stop("Invalid parameter values for lambda and mu!")
    # }

    # massExtinctionTimes=c(),
    # massExtinctionSurvivalProbabilities=c(),
    # missingSpecies = c(),
    # timesMissingSpecies = c(),
    # tess.likelihood 11
    # samplingProbability=1.0,
    # samplingStrategy="uniform",
    # MRCA=TRUE,
    # CONDITION="survival",
    birth_rate = kwargs.get("birth_rate")
    death_rate = kwargs.get("death_rate")
    massExtinctionTimes = None
    massExtinctionSurvivalProbabilities = None
    sampling_probability = kwargs.get("sampling_probability", 1.0)
    sampling_strategy = kwargs.get("sampling_strategy", "uniform")
    is_mrca_included = kwargs.get("is_mrca_included", True)
    condition_on = kwargs.get("condition_on", "survival")

    ntax = len(internal_node_ages) + 1
    PRESENT = max(internal_node_ages)
    times = [(PRESENT - t) for t in internal_node_ages]

    # if we condition on the MRCA, then we need to remove the root speciation event
    if is_mrca_included:
        times = times[1:]

    # set the uniform taxon sampling probability
    if sampling_strategy == "uniform":
        rho = sampling_probability
    else:
        rho = 1.0

    # initialize the log likelihood
    lnl = 0.0

    # what do we condition on?
    # did we condition on survival?
    if condition_on == "survival":
        lnl = - _p_survival_constant(birth_rate,death_rate,massExtinctionTimes,massExtinctionSurvivalProbabilities,rho,0,PRESENT,PRESENT)

    # death_rateltiply the probability of a descendant of the initial species
    lnl = lnl + _p1_constant(
            birth_rate=birth_rate,
            death_rate=death_rate,
            massExtinctionTimes=massExtinctionTimes,
            massExtinctionSurvivalProbabilities=massExtinctionSurvivalProbabilities,
            sampling_probability=rho,
            t=0,
            T=PRESENT)

    # add the survival of a second species if we condition on the MRCA
    if is_mrca_included:
        lnl = 2 * lnl

    # did we condition on observing n species today
    if condition_on == "taxa":
        lnl = lnl - _p_N_constant(
                birth_rate=birth_rate,
                death_rate=death_rate,
                massExtinctionTimes=massExtinctionTimes,
                massExtinctionSurvivalProbabilities=massExtinctionSurvivalProbabilities,
                sampling_probability=sampling_probability,
                i=ntax,
                s=0,
                t=PRESENT,
                SURVIVAL=False,
                MRCA=is_mrca_included)

    # if we assume diversified sampling, we need to death_rateltiply with the
    # probability that all missing species happened after the last speciation
    # event
    if sampling_strategy == "diversified":
        # We use equation (5) of Hoehna et al. "Inferring Speciation and Extinction Rates under Different Sampling Schemes"
        lastEvent = times[-1]
        p_0_T = 1.0 - math.exp(_p_survival_constant(birth_rate,death_rate,massExtinctionTimes,massExtinctionSurvivalProbabilities,1.0,0,PRESENT,PRESENT)) * math.exp((death_rate-birth_rate)*PRESENT)
        p_0_t = 1.0 - math.exp(_p_survival_constant(birth_rate,death_rate,massExtinctionTimes,massExtinctionSurvivalProbabilities,1.0,lastEvent,PRESENT,PRESENT)) * math.exp((death_rate-birth_rate)*(PRESENT-lastEvent))
        F_t = p_0_t / p_0_T
        # get an estimate of the actual number of taxa
        m = round(float(ntax) / sampling_probability)
        # remove the number of species that we started with
        if is_mrca_included:
            k = 2
        else:
            k = 1
        lnl = lnl + (m-ntax) * math.log(F_t) + math.log(combinatorics.choose(int(m-k),int(ntax-k)))

    # multiply the probability for the missing species
    # if len(missing_species) > 0:
    #     # We use equation (5) of Hoehna et al. "Inferring Speciation and Extinction Rates under Different Sampling Schemes"
    #     # now iterate over the vector of missing species per interval
    #     lastEvent = timesMissingSpecies
    #     p_0_T = 1.0 - math.exp( _p_survival_constant(birth_rate,death_rate,massExtinctionTimes,massExtinctionSurvivalProbabilities,1.0,0,PRESENT,PRESENT,log=TRUE) + ((death_rate-birth_rate)*PRESENT) )
    #     p_0_t = 1.0 - math.exp( _p_survival_constant(birth_rate,death_rate,massExtinctionTimes,massExtinctionSurvivalProbabilities,1.0,lastEvent,PRESENT,PRESENT,log=TRUE) + ((death_rate-birth_rate)*(PRESENT-lastEvent)) )
    #     log_F_t = math.log(p_0_t) - math.log(p_0_T)
    #     # get an estimate of the actual number of taxa
    #     m = missingSpecies
    #     # remove the number of species that we started with
    #     lnl = lnl + sum( m * log_F_t ) #+ lchoose(m-k,ntax-k)


    # multiply the probability for each speciation time
    if len(times) > 0:
        # lnl = lnl + len(times)*math.log(birth_rate) + sum(tess.equations.p1.constant(birth_rate,death_rate,massExtinctionTimes,massExtinctionSurvivalProbabilities,rho,times,PRESENT,log=TRUE))
        lnl += (len(times) * math.log(birth_rate))
        for time in times:
            lnl += _p1_constant(
                    birth_rate=birth_rate,
                    death_rate=death_rate,
                    massExtinctionTimes=massExtinctionTimes,
                    massExtinctionSurvivalProbabilities=massExtinctionSurvivalProbabilities,
                    sampling_probability=rho,
                    t=time,
                    T=PRESENT)

    return lnl



################################################################################
# From: TESS
#
#   https://github.com/hoehna/TESS
#
#   H{"o}hna S. 2013. Fast simulation of reconstructed phylogenies under
#   global time-dependent birth--death processes. Bioinformatics, 29(11)
#   1367-1374.
#
#   Copyright (c) 2012- Sebastian Hoehna
#
#   TESS is free software; you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
# @brief  Calculate the probability of n lineage existing at time t
#         if we started with 1 (or 2) lineage at time s.
#
#
# @see    Equation (3), (5) and (12) in Hoehna, S.: Fast simulation of reconstructed phylogenies
#         under global, time-dependent birth-death processes. 2013, Bioinformatics
#
# @date Last modified: 2013-01-30
# @author Sebastian Hoehna
# @version 1.1
# @since 2012-09-11, version 1.0
#
# @param    lambda                                        scalar        speciation rate
# @param    mu                                            scalar        extinction rate
# @param    massExtinctionTimes                           vector        timse at which mass-extinctions happen
# @param    massExtinctionSurvivalProbabilities           vector        survival probability of a mass extinction event
# @param    samplingProbability                           scalar        probability of uniform sampling at present
# @param    i                                             scalar        number of lineages
# @param    s                                             scalar        start time
# @param    t                                             scalar        stop time
# @param    MRCA                                          boolean       does the tree start at the mrca?
# @param    log                                           bool          if in log-scale
# @return                                                 scalar        log-probability
#
################################################################################
# tess.equations.pN.constant = function(lambda,mu,massExtinctionTimes,massExtinctionSurvivalProbabilities,samplingProbability,i,s,t,SURVIVAL=FALSE,MRCA=FALSE,log=FALSE) {
def _p_N_constant(
        birth_rate,
        death_rate,
        massExtinctionTimes,
        massExtinctionSurvivalProbabilities,
        sampling_probability,
        i,
        s,
        t,
        SURVIVAL=False,
        MRCA=False,):
    if i < 1: # we assume conditioning on survival
        p = 0
    elif i == 1:
        if MRCA: # we assume conditioning on survival of the two species
            p = 0
        else:
            if SURVIVAL:
                p =  _p_survival_constant(birth_rate,death_rate,massExtinctionTimes,massExtinctionSurvivalProbabilities,sampling_probability,s,t,t) + (death_rate-birth_rate)*(t-s) - math.log(sampling_probability)
            else:
                p =  2*_p_survival_constant(birth_rate,death_rate,massExtinctionTimes,massExtinctionSurvivalProbabilities,sampling_probability,s,t,t) + (death_rate-birth_rate)*(t-s) - math.log(sampling_probability)
    else:
        p_s = math.exp(_p_survival_constant(birth_rate,death_rate,massExtinctionTimes,massExtinctionSurvivalProbabilities,sampling_probability,s,t,t))
        r   = (death_rate-birth_rate)*(t-s) - math.log(sampling_probability)
        # for (j in seq_len(length(massExtinctionTimes)) ) {
        #     cond =  (s < massExtinctionTimes[j]) & (t >= massExtinctionTimes[j])
        #     r  = r - ifelse(cond, log(massExtinctionSurvivalProbabilities[j]), 0.0)
        e = p_s * math.exp(r)
        if e > 1:
            e = 1.0

        if not MRCA:
            if SURVIVAL:
                p   = math.log(p_s) + r + math.log( 1 - e) * (i-1)
            else:
                p   = 2*math.log(p_s) + r + math.log( 1 - e) * (i-1)
        else:
            if SURVIVAL:
                p = math.log(i-1) + 2*math.log(p_s) + 2*r + math.log( 1 - e) * (i-2)
            else:
                p = math.log(i-1) + 4*math.log(p_s) + 2*r + math.log( 1 - e) * (i-2)

    return p

################################################################################
# From: TESS
#
#   https://github.com/hoehna/TESS
#
#   H{"o}hna S. 2013. Fast simulation of reconstructed phylogenies under
#   global time-dependent birth--death processes. Bioinformatics, 29(11)
#   1367-1374.
#
#   Copyright (c) 2012- Sebastian Hoehna
#
#   TESS is free software; you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
# @brief  Calculate the probability of survival in the interval [t_low,t_high].
#
# See Equation (2) from Hoehna, S., 2013, Fast simulation of reconstructed phylogenies under global, time-dependent birth-death processes
#
#
# @date Last modified: 2013-01-30
# @author Sebastian Hoehna
# @version 1.1
# @since 2012-09-11, version 1.0
#
# @param    lambda                                        scalar        speciation rate
# @param    mu                                            scalar        extinction rate
# @param    massExtinctionTimes                           vector        timse at which mass-extinctions happen
# @param    massExtinctionSurvivalProbabilities           vector        survival probability of a mass extinction event
# @param    samplingProbability                           scalar        probability of uniform sampling at present
# @param    t_low                                         scalar        starting time
# @param    t_high                                        scalar        end time
# @param    T                                             scalar        present time (time goes forward and the origin/MRCA might be 0)
# @param    log                                           bool          if in log-scale
# @return                                                 scalar        probability of survival in [t,tau]
#
################################################################################
# tess.equations.pSurvival.constant = function(lambda,mu,massExtinctionTimes,massExtinctionSurvivalProbabilities,samplingProbability,t_low,t_high,T,log=FALSE) {
def _p_survival_constant(
        birth_rate,
        death_rate,
        massExtinctionTimes,
        massExtinctionSurvivalProbabilities,
        sampling_probability,
        t_low,
        t_high,
        T):

    # compute the rate
    rate = death_rate - birth_rate

    # do the integration of int_{t_low}^{t_high} ( death_rate(s) exp(rate(t,s)) ds )
    # where rate(t,s) = int_{t}^{s} ( death_rate(x)-birth_rate(x) dx ) - sum_{for all t < m_i < s in massExtinctionTimes }( log(massExtinctionSurvivalProbability[i]) )

    # we compute the integral stepwise for each epoch between mass-extinction events
    # add mass-extinction
    accudeath_ratelatedMassExtinction = 1.0
    prev_time = t_low
    den = 1.0
    # if ( length(massExtinctionTimes) > 0 ) {
    #     for (j in 1:length(massExtinctionTimes) ) {
    #         cond =  (t_low < massExtinctionTimes[j]) & (t_high >= massExtinctionTimes[j])
    #         # compute the integral for this time episode until the mass-extinction event
    #     #       den = den + ifelse(cond, exp(-rate*t_low) * death_rate / (rate * accudeath_ratelatedMassExtinction ) * ( exp(rate* massExtinctionTimes[j]) - exp(rate*prev_time)) , 0 )
    #         den = den + cond * exp(-rate*t_low) * death_rate / (rate * accudeath_ratelatedMassExtinction ) * ( exp(rate* massExtinctionTimes[j]) - exp(rate*prev_time))
    #         # store the current time so that we remember from which episode we need to integrate next
    #     #       prev_time = ifelse(cond, massExtinctionTimes[j], prev_time)
    #         prev_time[cond] = massExtinctionTimes[j]
    #         accudeath_ratelatedMassExtinction = accudeath_ratelatedMassExtinction * ifelse(cond, massExtinctionSurvivalProbabilities[j], 1.0)
    #         # integrate over the tiny time interval of the mass-extinction event itself and add it to the integral
    #     #       den = den - ifelse(cond, (massExtinctionSurvivalProbabilities[j]-1) / accudeath_ratelatedMassExtinction * exp( rate*(massExtinctionTimes[j] - t_low) ), 0.0 )
    #         den = den - cond * (massExtinctionSurvivalProbabilities[j]-1) / accudeath_ratelatedMassExtinction * exp( rate*(massExtinctionTimes[j] - t_low) )
    #     }
    # }

    # add the integral of the final epoch until the present time
    den = den + math.exp(-rate*t_low) * death_rate / (rate * accudeath_ratelatedMassExtinction ) * ( math.exp(rate*t_high) - math.exp(rate*prev_time))

    # add sampling
    if (sampling_probability < 1) and (t_low < T) and (t_high >= T):
        accudeath_ratelatedMassExtinction = accudeath_ratelatedMassExtinction *  sampling_probability
        cond = 1
    else:
        cond = 0

    #   den = den - ifelse(cond, (sampling_probability-1)*math.exp( rate*(T-t_low) ) / accudeath_ratelatedMassExtinction, 0.0)
    den = den - cond * (sampling_probability-1)*math.exp( rate*(T-t_low) ) / accudeath_ratelatedMassExtinction

    res = 1.0 / den
    res = math.log(res)

    return (res)


################################################################################
# From: TESS
#
#   https://github.com/hoehna/TESS
#
#   H{"o}hna S. 2013. Fast simulation of reconstructed phylogenies under
#   global time-dependent birth--death processes. Bioinformatics, 29(11)
#   1367-1374.
#
#   Copyright (c) 2012- Sebastian Hoehna
#
#   TESS is free software; you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
# @brief  Calculate the probability of exactly 1 lineage surviving until time T
#         if we started with 1 lineage at time t.
#
# @see    Equation (4) in Hoehna, S.: Fast simulation of reconstructed phylogenies
#         under global, time-dependent birth-death processes. 2013, Bioinformatics
#
# @date Last modified: 2013-01-30
# @author Sebastian Hoehna
# @version 1.1
# @since 2012-09-11, version 1.0
#
# @param    lambda                                        scalar        speciation rate
# @param    mu                                            scalar        extinction rate
# @param    massExtinctionTimes                           vector        timse at which mass-extinctions happen
# @param    massExtinctionSurvivalProbabilities           vector        survival probability of a mass extinction event
# @param    samplingProbability                           scalar        probability of uniform sampling at present
# @param    t                                             scalar        time
# @param    T                                             scalar        present time
# @param    log                                           bool          if in log-scale
# @return                                                 scalar        ln-probability
#
################################################################################
# tess.equations.p1.constant = function(lambda,mu,massExtinctionTimes,massExtinctionSurvivalProbabilities,samplingProbability,t,T,log=FALSE) {
def _p1_constant(
        birth_rate,
        death_rate,
        massExtinctionTimes,
        massExtinctionSurvivalProbabilities,
        sampling_probability,
        t,
        T):
    # compute the survival probability
    a = _p_survival_constant(
            birth_rate=birth_rate,
            death_rate=death_rate,
            massExtinctionTimes=massExtinctionTimes,
            massExtinctionSurvivalProbabilities=massExtinctionSurvivalProbabilities,
            sampling_probability=sampling_probability,
            t_low=t,
            t_high=T,
            T=T)
    # compute the rate
    rate = (death_rate - birth_rate)*(T-t)
    # add mass-extinction
    # for (j in range(len(massExtinctionTimes)) ) :
    #     rate = rate - ifelse( t < massExtinctionTimes[j] & T >= massExtinctionTimes[j], log(massExtinctionSurvivalProbabilities[j]), 0 )
    # add sampling
    rate = rate - math.log(sampling_probability)
    p = 2*a + rate
    return p



#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Handling deprecation warnings and messages correctly.
"""

import os
import warnings
from dendropy.utility import metavar

DEPRECATION_WARNING_FILTER = None
_DEPRECATION_WARNINGS_CONFIGURED = False

class CriticalDeprecationWarning(UserWarning):
    pass

def configure_deprecation_warning_behavior(warning_filter=None):
    global DEPRECATION_WARNING_FILTER
    global _DEPRECATION_WARNINGS_CONFIGURED
    if warning_filter is None:
        warning_filter = os.environ.get(metavar.DEPRECATION_WARNING_FILTER, "default")
    DEPRECATION_WARNING_FILTER = warning_filter
    warnings.simplefilter(DEPRECATION_WARNING_FILTER,
            CriticalDeprecationWarning)
    _DEPRECATION_WARNINGS_CONFIGURED = True

def _initialize_deprecation_warnings():
    global _DEPRECATION_WARNINGS_CONFIGURED
    if not _DEPRECATION_WARNINGS_CONFIGURED:
        configure_deprecation_warning_behavior()

def dendropy_deprecation_warning(**kwargs):
    _initialize_deprecation_warnings()
    leader = "  # "
    stacklevel = kwargs.pop("stacklevel", 3)
    if "message" in kwargs:
        message = kwargs["message"]
    elif "old_construct" in kwargs or "new_construct" in kwargs:
        message = []
        message.append("")
        if "preamble" in kwargs:
            message.append(leader + kwargs["preamble"])
        message.append(leader + "Instead of:")
        for construct in kwargs["old_construct"].split("\n"):
            message.append(leader + "    {}".format(construct))
        message.append(leader + "Use:")
        for construct in kwargs["new_construct"].split("\n"):
            message.append(leader + "    {}".format(construct))
        if "epilog" in kwargs:
            message.append(leader + kwargs["epilog"])
        message = "\n".join(message)
    _initialize_deprecation_warnings()
    old_formatwarning = warnings.formatwarning
    warnings.warn(
            message=message,
            category=CriticalDeprecationWarning,
            stacklevel=stacklevel,
            )



#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Deserialization of NeXML-formatted data.
"""

import re
import collections
from dendropy.dataio import ioservice
from dendropy.utility import container
from dendropy.utility import textprocessing
from dendropy.utility import error
from dendropy.dataio import xmlprocessing
from dendropy.datamodel.charstatemodel import StateAlphabet, StateIdentity

SUPPORTED_NEXML_NAMESPACES = ('http://www.nexml.org/1.0', 'http://www.nexml.org/2009')

def _from_nexml_tree_length_type(type_attr):
    """
    Given an attribute string read from a nexml tree element, returns
    the python class of the edge length attribute.
    """
    if type_attr == "nex:IntTree":
        return int
    else:
        return float

class _AnnotationParser(object):

    def __init__(self, namespace_registry=None):
        self._namespace_registry = namespace_registry

    def _parse_annotations(self, annotated, nxelement):
        attrib = nxelement.attrib
        xml_type = nxelement.parse_type()
        if xml_type == 'LiteralMeta':
            value = attrib.get("content", None)
            key = attrib.get("property", None)
            annotate_as_reference = False
        else:
            value = attrib.get("href", None)
            key = attrib.get("rel", None)
            annotate_as_reference = True
        datatype_hint = attrib.get("datatype", None)
        if key is None:
            raise ValueError("Could not determine property/rel for meta element: %s\n%s" % (nxelement, attrib))
        name_prefix, name = textprocessing.parse_curie_standard_qualified_name(key)
        try:
            namespace = self._namespace_registry.prefix_namespace_map[name_prefix]
        except KeyError:
            raise ValueError("CURIE-standard prefix '%s' not defined in document: %s" % (name_prefix, self._namespace_registry))
        if datatype_hint is not None:
            dt_prefix, dt = textprocessing.parse_curie_standard_qualified_name(datatype_hint)
            if dt_prefix is not None:
                try:
                    dt_namespace = self._namespace_registry.prefix_namespace_map[dt_prefix]
                except KeyError:
                    raise ValueError("CURIE-standard prefix '%s' not defined in document: %s" % (dt_prefix, self._namespace_registry))
                if dt_namespace.startswith("http://www.w3.org/2001/XMLSchema"):
                    value = self._coerce_to_xml_schema_type(value, dt)
                elif dt_namespace.startswith("http://www.nexml.org/1.0") or dt_namespace.startswith("http://www.nexml.org/2009"):
                    value = self._coerce_to_nexml_type(value, dt)
                elif dt_namespace.startswith("http://dendropy.org") or dt_namespace.startswith("http://packages.python.org/DendroPy") or dt_namespace.startswith("http://pypi.org/project/DendroPy/"):
                    value = self._coerce_to_dendropy_type(value, dt)
        a = annotated.annotations.add_new(
                name=name,
                value=value,
                datatype_hint=datatype_hint,
                name_prefix=name_prefix,
                namespace=namespace,
                name_is_prefixed=False,
                annotate_as_reference=annotate_as_reference)
        top_annotations = [i for i in nxelement.findall_annotations()]
        for annotation in top_annotations:
            self._parse_annotations(a, annotation)

    def _coerce_to_xml_schema_type(self, value, type_name):
        if type_name in ("boolean"):
            if value.lower() in ("1", "t", "true", "y", "yes",):
                return True
            else:
                return False
        elif type_name in ("decimal", "float", "double"):
            coerce_type = float
        elif type_name in ("byte", "int", "integer", "long", "negativeInteger", "nonNegativeInteger", "nonPositiveInteger", "short", "unsignedInt", "unsignedLong", "unsignedShort"):
            coerce_type = int
        else:
            return value
        return self._coerce_type(value, coerce_type)

    def _coerce_to_dendropy_type(self, value, type_name):
        if type_name in ("decimalRange"):
            try:
                value = [float(v) for v in value.split()]
            except KeyError:
                pass
        return value

    def _coerce_to_nexml_type(self, value, type_name):
        if type_name in ("ContinuousSeq"):
            try:
                value = [float(v) for v in value.split()]
            except KeyError:
                pass
        return value

    def _coerce_type(self, value, to_type):
        try:
            value = to_type(value)
        except ValueError:
            pass
        return value


class NexmlElement(xmlprocessing.XmlElement):

    def __init__(self, element, default_namespace=None):
        # if default_namespace is None:
        #     default_namespace = NexmlReader.DEFAULT_NEXML_NAMESPACE
        # else:
        #     default_namespace = default_namespace
        xmlprocessing.XmlElement.__init__(self,
                element=element,
                default_namespace=default_namespace)
        self.type_parse_pattern = re.compile(r"([A-Za-z0-9]+?):(.+)")

    ## Annotations ##

    def findall_annotations(self):
        return self.namespaced_findall("meta")

    ## TaxonSet Elements ##

    def iter_otus(self):
        return self.namespaced_getiterator("otus")

    def findall_otu(self):
        return self.namespaced_findall("otu")

    ## Characters ##

    def iter_characters(self):
        return self.namespaced_getiterator("characters")

    def find_char_format(self):
        return self.namespaced_find("format")

    def find_char_matrix(self):
        return self.namespaced_find("matrix")

    def findall_multistate_member(self):
        return self.namespaced_findall("member")

    def findall_polymorphic_state_set(self):
        return self.namespaced_findall("polymorphic_state_set")

    def findall_uncertain_state_set(self):
        return self.namespaced_findall("uncertain_state_set")

    def findall_char_state(self):
        return self.namespaced_findall("state")

    def findall_char_states(self):
        return self.namespaced_findall("states")

    def findall_char(self):
        return self.namespaced_findall("char")

    def findall_char_row(self):
        return self.namespaced_findall("row")

    def findall_char_cell(self):
        return self.namespaced_findall("cell")

    def find_char_seq(self):
        return self.namespaced_findtext("seq")

    ## Trees ##

    def iter_trees(self):
        return self.namespaced_getiterator("trees")

    def findall_tree(self):
        return self.namespaced_findall("tree")

    def findall_node(self):
        return self.namespaced_findall("node")

    def findall_edge(self):
        return self.namespaced_findall("edge")

    def find_rootedge(self):
        return self.namespaced_find("rootedge")

    def parse_type(self):
        type_value = self._element.get('{http://www.w3.org/2001/XMLSchema-instance}type', None)
        if type_value is None:
            raise ValueError("Type not specified for element '%s'" % self._element.get("id", None))
        m = self.type_parse_pattern.match(type_value)
        if m:
            return m.groups()[1]
        else:
            return type_value

class NexmlReader(ioservice.DataReader, _AnnotationParser):
    "Implements thinterface for handling NEXML files."

    ###########################################################################
    ## Life-cycle and Setup

    DEFAULT_NEXML_NAMESPACE = "http://www.nexml.org/2009"

    def __init__(self, **kwargs):
        """
        Keyword Arguments
        -----------------

        default_namespace : str
            Default namespace to use for elements.
        case_sensitive_taxon_labels: boolean, default: |False|
            If |True|, then case is respected when matching taxon names.
            Default is |False|: case is ignored.
        ignore_unrecognized_keyword_arguments : boolean, default: |False|
            If |True|, then unsupported or unrecognized keyword arguments will
            not result in an error. Default is |False|: unsupported keyword
            arguments will result in an error.

        """
        _AnnotationParser.__init__(self)
        ioservice.DataReader.__init__(self)
        self.default_namespace = kwargs.pop("default_namespace", NexmlReader.DEFAULT_NEXML_NAMESPACE)
        self.case_sensitive_taxon_labels = kwargs.pop('case_sensitive_taxon_labels', False)

        ### NOTE: following are not actually supported.
        ### They are here because some tests automatically add include them on calls.
        ### TODO: remove these keywords from generic tests
        self.suppress_internal_node_taxa = kwargs.pop("suppress_internal_node_taxa", True)
        self.suppress_leaf_node_taxa = kwargs.pop("suppress_external_node_taxa", False) # legacy (will be deprecated)
        self.suppress_leaf_node_taxa = kwargs.pop("suppress_leaf_node_taxa", self.suppress_leaf_node_taxa)
        self.data_type = kwargs.pop("data_type", None)

        self.check_for_unused_keyword_arguments(kwargs)

        # Set up parsing meta-variables
        self._id_taxon_namespace_map = {}
        self._id_taxon_map = {}
        self._taxon_namespace_factory = None
        self._tree_list_factory = None
        self._char_matrix_factory = None
        self._state_alphabet_factory = None
        self._global_annotations_target = None
        self._taxon_namespaces = []
        self._char_matrices = []
        self._tree_lists = []
        self._product = None

    ###########################################################################
    ## Reader Interface

    def _read(self,
            stream,
            taxon_namespace_factory=None,
            tree_list_factory=None,
            char_matrix_factory=None,
            state_alphabet_factory=None,
            global_annotations_target=None):
        xml_doc = xmlprocessing.XmlDocument(file_obj=stream,
                subelement_factory=self._subelement_factory)
        self._namespace_registry = xml_doc.namespace_registry
        self._taxon_namespace_factory = taxon_namespace_factory
        self._tree_list_factory = tree_list_factory
        self._char_matrix_factory = char_matrix_factory
        self._state_alphabet_factory = state_alphabet_factory
        self._global_annotations_target = global_annotations_target
        self._parse_document(xml_doc)
        self._product = self.Product(
                taxon_namespaces=self._taxon_namespaces,
                tree_lists=self._tree_lists,
                char_matrices=self._char_matrices)
        return self._product

    ###########################################################################
    ## Support

    def _subelement_factory(self, element):
        return NexmlElement(element, default_namespace=self.default_namespace)

    ###########################################################################
    ## Data Management

    def _new_taxon_namespace(self, label=None):
        if self.attached_taxon_namespace is not None:
            return self.attached_taxon_namespace
        taxon_namespace = self._taxon_namespace_factory(label=label)
        self._taxon_namespaces.append(taxon_namespace)
        return taxon_namespace

    def _new_char_matrix(self, data_type, taxon_namespace, label=None, **kwargs):
        char_matrix = self._char_matrix_factory(
                data_type,
                taxon_namespace=taxon_namespace,
                label=label,
                **kwargs)
        self._char_matrices.append(char_matrix)
        return char_matrix

    def _new_state_alphabet(self, *args, **kwargs):
        return self._state_alphabet_factory(*args, **kwargs)

    def _new_tree_list(self, taxon_namespace, label=None):
        tree_list = self._tree_list_factory(
                taxon_namespace=taxon_namespace,
                label=label)
        self._tree_lists.append(tree_list)
        return tree_list

    ###########################################################################
    ## Parsers

    ## Following methods are class-specific ###

    def _parse_document(self, xml_doc):
        xml_root = xml_doc.root
        self._parse_taxon_namespaces(xml_root)
        if self._char_matrix_factory is not None:
            self._parse_char_matrices(xml_root)
        if self._tree_list_factory is not None:
            self._parse_tree_lists(xml_root)
        if self._global_annotations_target is not None:
            top_annotations = [i for i in xml_root.findall_annotations()]
            for annotation in top_annotations:
                self._parse_annotations(self._global_annotations_target, annotation)

    def _parse_taxon_namespaces(self, xml_root):
        for nxtaxa in xml_root.iter_otus():
            taxon_namespace_label = nxtaxa.get('label', None)
            taxon_namespace = self._new_taxon_namespace(label=taxon_namespace_label)
            taxon_namespace_id = nxtaxa.get('id', id(taxon_namespace))
            self._id_taxon_namespace_map[taxon_namespace_id] = taxon_namespace
            annotations = [i for i in nxtaxa.findall_annotations()]
            for annotation in annotations:
                self._parse_annotations(taxon_namespace, annotation)
            if self.case_sensitive_taxon_labels:
                label_taxon_map = {}
            else:
                label_taxon_map = container.OrderedCaselessDict()
            if self.attached_taxon_namespace is not None:
                for t in taxon_namespace:
                    label_taxon_map[t.label] = t
            for idx, nxtaxon in enumerate(nxtaxa.findall_otu()):
                taxon = None
                taxon_label = nxtaxon.get('label', None)
                taxon_oid = nxtaxon.get('id', id(nxtaxon))
                if taxon_label is not None and self.attached_taxon_namespace is not None:
                    # taxon = label_taxon_map.get_taxon(
                    #         label=taxon_label,
                    #         case_sensitive=self.case_sensitive_taxon_labels)
                    try:
                        taxon = label_taxon_map[taxon_label]
                    except KeyError:
                        taxon = None
                if taxon is None:
                    taxon = taxon_namespace.new_taxon(label=taxon_label)
                annotations = [i for i in nxtaxon.findall_annotations()]
                for annotation in annotations:
                    self._parse_annotations(taxon, annotation)
                self._id_taxon_map[(taxon_namespace_id, taxon_oid)] = taxon

    def _parse_char_matrices(self, xml_root):
        nxc = _NexmlCharBlockParser(self._namespace_registry,
                self._id_taxon_namespace_map,
                self._id_taxon_map,
                self._new_char_matrix,
                self._state_alphabet_factory)
        for char_matrix_element in xml_root.iter_characters():
            nxc.parse_char_matrix(char_matrix_element)

    def _parse_tree_lists(self, xml_root):
        for trees_idx, trees_element in enumerate(xml_root.iter_trees()):
            self._parse_tree_list(trees_element, trees_idx, True)

    def _parse_tree_list(self, nxtrees, trees_idx=None, add_to_tree_list=True):
        trees_id = nxtrees.get('id', "Trees" + str(trees_idx))
        trees_label = nxtrees.get('label', None)
        otus_id = nxtrees.get('otus', None)
        if otus_id is None:
            raise Exception("Taxa block not specified for trees block '{}'".format(otus_id))
        taxon_namespace = self._id_taxon_namespace_map.get(otus_id, None)
        if not taxon_namespace:
            raise Exception("Tree block '{}': Taxa block '{}' not found".format(trees_id, otus_id))
        tree_list = self._new_tree_list(
                label=trees_label,
                taxon_namespace=taxon_namespace)
        annotations = [i for i in nxtrees.findall_annotations()]
        for annotation in annotations:
            self._parse_annotations(tree_list, annotation)
        tree_parser = _NexmlTreeParser(
                id_taxon_map=self._id_taxon_map,
                annotations_processor_fn=self._parse_annotations,
                )
        for tree_element in nxtrees.findall_tree():
            tree_obj = tree_list.new_tree()
            tree_parser.build_tree(tree_obj, tree_element, otus_id)

class _NexmlTreeParser(object):

    EdgeInfo = collections.namedtuple(
            "EdgeInfo",
            ["edge_id", "label", "tail_node_id", "head_node_id", "length", "annotations"]
            )

    def __init__(self,
            id_taxon_map,
            annotations_processor_fn,
            ):
        self._id_taxon_map = id_taxon_map
        self._process_annotations = annotations_processor_fn

    def build_tree(self, tree_obj, tree_element, otus_id):
        tree_obj.label = tree_element.get('label', '')
        tree_type_attr = tree_element.get('{http://www.w3.org/2001/XMLSchema-instance}type')
        tree_obj.length_type = _from_nexml_tree_length_type(tree_type_attr)
        tree_obj.is_rooted = False # default unless explicitly set
        annotations = [i for i in tree_element.findall_annotations()]
        for annotation in annotations:
            self._process_annotations(tree_obj, annotation)
        unparented_node_set, node_id_map, root_node_list = self._parse_nodes(
                tree_element,
                tree_obj.node_factory,
                otus_id)
        if len(root_node_list) > 1:
            raise Exception("Multiple root nodes defined but the DendroPy tree model currently only supported single-root trees")
        elif len(root_node_list) == 1:
            tree_obj.seed_node = root_node_list[0]
            tree_obj.is_rooted = True
        edges_info = self._parse_edges_info(tree_element, length_type=tree_obj.length_type)
        for edge_info in edges_info:
            if edge_info.head_node_id is None:
                raise Exception("Edge '{}' does not specify a head (target) node".format(edge_info.edge_id, edge_info.head_node_id))
            try:
                head_node = node_id_map[edge_info.head_node_id]
            except KeyError:
                raise Exception("Edge '{}' specifies a non-defined head (target) node '{}'".format(edge_info.edge_id, edge_info.head_node_id))
            if edge_info.tail_node_id is not None:
                try:
                    tail_node = node_id_map[edge_info.tail_node_id]
                except KeyError:
                    raise Exception("Edge '{}' specifies a non-defined tail (source) node '{}'".format(edge_info.edge_id, edge_info.tail_node_id))
                head_node.parent_node = tail_node
                assert head_node.edge.tail_node is tail_node
                unparented_node_set.remove(head_node)
            else:
                assert head_node.parent_node is None
            self._set_edge_details(head_node.edge, edge_info)

        # find node(s) without parent
        # unparented_node_sets = []
        # for node in nodes.values():
        #     if node.parent_node == None:
        #         unparented_node_sets.append(node)

        # If one unparented_node_sets node found, this is the root: we use
        # it as the tree head node. If multiple unparented_node_sets nodes
        # are found, then we add them all as child_nodes of the
        # existing head node. If none, then we have some sort of
        # cyclicity, and we are not dealing with a tree.
        if len(unparented_node_set) == 1:
            unparented_node = unparented_node_set.pop()
            if root_node_list:
                if root_node_list[0] is not unparented_node:
                    raise Exception("Tree already has an explictly defined root node, but node without parent found: {}".format(unparented_node))
            else:
                tree_obj.seed_node = unparented_node
        elif len(unparented_node_set) > 1:
            for node in unparented_node_set:
                tree_obj.seed_node.add_child(node)
        else:
            raise Exception("Structural error: tree must be acyclic.")

        root_edge_element = tree_element.find_rootedge()
        if root_edge_element is not None:
            root_edge_info = self._parse_edge_info(root_edge_element, tree_obj.length_type)
            root_node = node_id_map[root_edge_info.head_node_id]
            if root_node is not tree_obj.seed_node:
                raise Exception("Root edge does not subtend root node")
            root_edge = root_node.edge
            self._set_edge_details(root_edge, root_edge_info)
            # tree_obj.is_rooted = True

        return tree_obj

    def _parse_nodes(self, tree_element, node_factory, otus_id):
        """
        Given an XmlElement representation of a NEXML tree element,
        (`nex:tree`) this will return a dictionary of DendroPy Node
        objects with the node_id as the key.
        """
        node_set = set()
        node_id_map = {}
        root_node_list = []
        for nxnode in tree_element.findall_node():
            node_id = nxnode.get('id', None)
            node = node_factory()
            node_id_map[node_id] = node
            node_set.add(node)
            node_id_map[node_id].label = nxnode.get('label', None)
            taxon_id = nxnode.get('otu', None)
            if taxon_id is not None:
                try:
                    taxon = self._id_taxon_map[(otus_id, taxon_id)]
                except KeyError:
                    raise Exception("Taxon with id '{}' not defined '{}'".format(taxon_id, otus_id))
                node_id_map[node_id].taxon = taxon
            annotations = [i for i in nxnode.findall_annotations()]
            for annotation in annotations:
                self._process_annotations(node_id_map[node_id], annotation)
            rooting_state = nxnode.get('root', None)
            if rooting_state is not None and rooting_state.lower() in ("1", "t", "true"):
                root_node_list.append(node)
        return node_set, node_id_map, root_node_list

    def _parse_edges_info(self, tree_element, length_type):
        edges_info = []
        for nxedge in tree_element.findall_edge():
            edges_info.append(self._parse_edge_info(nxedge, length_type))
        return edges_info

    def _parse_edge_info(self, nxedge, length_type):
        edge_id = nxedge.get('id', None)
        edge_label = nxedge.get('label', None)
        tail_node_id = nxedge.get('source', None)
        # assert tail_node_id is not None
        head_node_id = nxedge.get('target', None)
        # assert head_node_id is not None
        edge_length_str = nxedge.get('length', 0.0)
        edge_length = None
        try:
            edge_length = length_type(edge_length_str)
        except ValueError:
            msg = "Edge {} 'length' attribute is not of type {}: '{}'".format(
                    edge_id, str(length_type), edge_length_str)
            raise Exception(msg)
        annotations = [i for i in nxedge.findall_annotations()]
        e = _NexmlTreeParser.EdgeInfo(
                edge_id=edge_id,
                label=edge_label,
                tail_node_id=tail_node_id,
                head_node_id=head_node_id,
                length=edge_length,
                annotations=annotations)
        return e

    def _set_edge_details(self, edge, edge_info):
        edge.length = edge_info.length
        edge.label = edge_info.label
        for annotation in edge_info.annotations:
            self._process_annotations(edge, annotation)

class _NexmlCharBlockParser(_AnnotationParser):
    "Parses an XmlElement representation of NEXML taxa blocks."

    def __init__(self,
            namespace_registry,
            id_taxon_namespace_map,
            id_taxon_map,
            char_matrix_factory,
            state_alphabet_factory,
            ):
        _AnnotationParser.__init__(self, namespace_registry)
        self._id_taxon_namespace_map = id_taxon_namespace_map
        self._id_taxon_map = id_taxon_map
        self._char_matrix_factory = char_matrix_factory
        self._state_alphabet_factory = state_alphabet_factory

        self._id_state_alphabet_map = {}
        self._id_state_map = {}
        self._id_chartype_map = {}
        self._char_types = []
        self._chartype_id_to_pos_map = {}

    def parse_char_matrix(self, nxchars):
        """
        Given an XmlElement representing a nexml characters block, this
        instantiates and returns a corresponding DendroPy CharacterMatrix object.
        """

        # clear
        self._id_state_alphabet_map = {}
        self._id_state_map = {}
        self._id_chartype_map = {}
        self._char_types = []
        self._chartype_id_to_pos_map = {}

        # initiaiize
        label = nxchars.get('label', None)
        char_matrix_oid = nxchars.get('oid', '')

        # set up taxa
        otus_id = nxchars.get('otus', None)
        if otus_id is None:
            raise Exception("Character Block %s (\"%s\"): Taxon namespace not specified" % (char_matrix_oid, label))
        taxon_namespace = self._id_taxon_namespace_map.get(otus_id, None)
        if not taxon_namespace:
            raise Exception("Character Block %s (\"%s\"): Specified taxon namespace not found" % (char_matrix_oid, label))

        # character matrix instantiation
        nxchartype = nxchars.parse_type()
        extra_kwargs = {}
        if nxchartype.startswith('Dna'):
            data_type = "dna"
        elif nxchartype.startswith('Rna'):
            data_type = "rna"
        elif nxchartype.startswith('Protein'):
            data_type = "protein"
        elif nxchartype.startswith('Restriction'):
            data_type = "restriction"
        elif nxchartype.startswith('Standard'):
            data_type = "standard"
            extra_kwargs["default_state_alphabet"] = None
        elif nxchartype.startswith('Continuous'):
            data_type = "continuous"
        else:
            raise Exception("Character Block %s (\"%s\"): Character type '%s' not supported" % (char_matrix_oid, label, nxchartype))
        char_matrix = self._char_matrix_factory(
                data_type,
                taxon_namespace=taxon_namespace,
                label=label,
                **extra_kwargs)

        # annotation processing
        annotations = [i for i in nxchars.findall_annotations()]
        for annotation in annotations:
            self._parse_annotations(char_matrix, annotation)

        # get state mappings
        nxformat = nxchars.find_char_format()
        if nxformat is not None:
            self.parse_characters_format(nxformat, data_type, char_matrix)
        elif data_type == "standard":
            self.create_standard_character_alphabet(char_matrix)

        nxmatrix = nxchars.find_char_matrix()
        annotations = [i for i in nxmatrix.findall_annotations()]
        for annotation in annotations:
            self._parse_annotations(char_matrix.taxon_seq_map, annotation)
        for nxrow in nxmatrix.findall_char_row():
            row_id = nxrow.get('id', None)
            label = nxrow.get('label', None)
            taxon_id = nxrow.get('otu', None)
            try:
                taxon = self._id_taxon_map[(otus_id, taxon_id)]
            except KeyError:
                raise error.DataParseError(message='Character Block %s (\"%s\"): Taxon with id "%s" not defined in taxa block "%s"' % (char_matrix.oid, char_matrix.label, taxon_id, otus_id))

            character_vector = char_matrix.new_sequence(taxon=taxon)
            annotations = [i for i in nxrow.findall_annotations()]
            for annotation in annotations:
                self._parse_annotations(character_vector, annotation)

            if data_type == "continuous":
                if nxchartype.endswith('Seqs'):
                    seq = nxrow.find_char_seq()
                    if seq is not None:
                        seq = seq.replace('\n\r', ' ').replace('\r\n', ' ').replace('\n', ' ').replace('\r',' ')
                        col_idx = -1
                        for char in seq.split(' '):
                            char = char.strip()
                            if char:
                                col_idx += 1
                                if len(self._char_types) <= col_idx:
                                    raise error.DataParseError(message="Character column/type ('<char>') not defined for character in position"\
                                        + " %d (matrix = '%s' row='%s', taxon='%s')" % (col_idx+1, char_matrix.oid, row_id, taxon.label))
                                character_vector.append(character_value=float(char), character_type=self._char_types[col_idx])
                else:
                    for nxcell in nxrow.findall_char_cell():
                        chartype_id = nxcell.get('char', None)
                        if chartype_id is None:
                            raise error.DataParseError(message="'char' attribute missing for cell: cell markup must indicate character column type for character"\
                                        + " (matrix = '%s' row='%s', taxon='%s')" % (char_matrix.oid, row_id, taxon.label))
                        if chartype_id not in self._id_chartype_map:
                            raise error.DataParseError(message="Character type ('<char>') with id '%s' referenced but not found for character" % chartype_id \
                                        + " (matrix = '%s' row='%s', taxon='%s')" % (char_matrix.oid, row_id, taxon.label))
                        chartype = self._id_chartype_map[chartype_id]
                        pos_idx = self._char_types.index(chartype)
#                         column = id_chartype_map[chartype_id]
#                         state = column.state_id_map[cell.get('state', None)]
                        # annotations = [i for i in nxcell.findall_annotations]
                        # for annotation in annotations:
                        #     self._parse_annotations(cell, annotation)
                        character_vector.append(character_value=float(nxcell.get('state')),
                                character_type=chartype)
            else:
                if nxchartype.endswith('Seqs'):
                    seq = nxrow.find_char_seq()
                    if seq is not None:
                        seq = seq.replace(' ', '').replace('\n', '').replace('\r', '')
                        col_idx = -1
                        for char in seq:
                            col_idx += 1
                            state_alphabet = char_matrix.character_types[col_idx].state_alphabet
                            try:
                                state = state_alphabet[char]
                            except KeyError:
                                raise error.DataParseError(message="Character Block row '%s', character position %s: State with symbol '%s' in sequence '%s' not defined" \
                                        % (row_id, col_idx, char, seq))
                            if len(self._char_types) <= col_idx:
                                raise error.DataParseError(message="Character column/type ('<char>') not defined for character in position"\
                                    + " %d (row='%s', taxon='%s')" % (col_idx+1, row_id, taxon.label))
                            character_type = self._char_types[col_idx]
                            character_vector.append(character_value=state,
                                    character_type=character_type)
                else:
                    for nxcell in nxrow.findall_char_cell():
                        chartype_id = nxcell.get('char', None)
                        if chartype_id is None:
                            raise error.DataParseError(message="'char' attribute missing for cell: cell markup must indicate character column type for character"\
                                        + " (matrix = '%s' row='%s', taxon='%s')" % (char_matrix_oid, row_id, taxon.label))
                        if chartype_id not in self._id_chartype_map:
                            raise error.DataParseError(message="Character type ('<char>') with id '%s' referenced but not found for character" % chartype_id \
                                        + " (matrix = '%s' row='%s', taxon='%s')" % (char_matrix_oid, row_id, taxon.label))
                        chartype = self._id_chartype_map[chartype_id]
                        state_alphabet = self._id_chartype_map[chartype_id].state_alphabet
                        pos_idx = self._chartype_id_to_pos_map[chartype_id]
                        state = self._id_state_map[ (state_alphabet, nxcell.get('state', None)) ]
                        character_vector.set_at(pos_idx,
                                character_value=state,
                                character_type=chartype)
                        # self._id_state_alphabet_map = {}
                        # self._id_state_map = {}
                        # self._id_chartype_map = {}

            char_matrix[taxon] = character_vector

        # if fixed_state_alphabet:
        #     char_matrix.remap_to_default_state_alphabet_by_symbol(purge_other_state_alphabets=True)

    def parse_ambiguous_state(self, nxstate, state_alphabet):
        """
        Parses an XmlElement represent an ambiguous discrete character state,
        ("uncertain_state_set")
        and returns a corresponding StateAlphabetElement object.
        """
        state_oid = nxstate.get('id', None)
        state_symbol = nxstate.get('symbol', None)
        token = nxstate.get('token', None)
        member_states = []
        for nxmember in nxstate.findall_multistate_member():
            member_state_id = nxmember.get('state', None)
            member_state = self._id_state_map[ (state_alphabet, member_state_id) ]
            member_states.append(member_state)
        state = state_alphabet.new_multistate(symbol=state_symbol,
                state_denomination=state_alphabet.AMBIGUOUS_STATE,
                member_states=member_states)
        assert (state_alphabet, state_oid) not in self._id_state_map
        self._id_state_map[ (state_alphabet, state_oid) ] = state
        if token is not None:
            state_alphabet.new_symbol_synonym(token, state_symbol)
        return state

    def parse_polymorphic_state(self, nxstate, state_alphabet):
        """
        Parses an XmlElement represent a polymorphic discrete character state,
        ("polymorphic_state_set")
        and returns a corresponding StateAlphabetElement object.
        """
        state_oid = nxstate.get('id', None)
        state_symbol = nxstate.get('symbol', None)
        token = nxstate.get('token', None)
        member_states = []
        for nxmember in nxstate.findall_multistate_member():
            member_state_id = nxmember.get('state', None)
            member_state = self._id_state_map[ (state_alphabet, member_state_id) ]
            member_states.append(member_state)
        for nxambiguous in nxstate.findall_uncertain_state_set():
            member_states.append(self.parse_ambiguous_state(nxstate, state_alphabet))
        state = state_alphabet.new_multistate(symbol=state_symbol,
                state_denomination=state_alphabet.POLYMORPHIC_STATE,
                member_states=member_states)
        assert (state_alphabet, state_oid) not in self._id_state_map
        self._id_state_map[ (state_alphabet, state_oid) ] = state
        if token is not None:
            state_alphabet.new_symbol_synonym(token, state_symbol)
        return state

    def parse_state_alphabet(self, nxstates):
        """
        Given an XmlElement representing a nexml definition of (discrete or standard) states
        ("states"), this returns a corresponding StateAlphabet object.
        """
        state_alphabet_oid = nxstates.get("id", None)
        state_alphabet = self._state_alphabet_factory()
        state_alphabet.autocompile_lookup_tables = False
        self._id_state_alphabet_map[state_alphabet_oid] = state_alphabet
        for nxstate in nxstates.findall_char_state():
            state_oid = nxstate.get('id', None)
            state_symbol = nxstate.get('symbol', None)
            state = state_alphabet.new_fundamental_state(symbol=state_symbol)
            token = nxstate.get('token', None)
            assert (state_alphabet, state_oid) not in self._id_state_map
            self._id_state_map[ (state_alphabet, state_oid) ] = state
            if token is not None:
                state_alphabet.new_symbol_synonym(token, state_symbol)
        for nxstate in nxstates.findall_uncertain_state_set():
            self.parse_ambiguous_state(nxstate, state_alphabet)
        for nxstate in nxstates.findall_polymorphic_state_set():
            self.parse_polymorphic_state(nxstate, state_alphabet)
        state_alphabet.autocompile_lookup_tables = True
        state_alphabet.compile_lookup_mappings()
        return state_alphabet

    def parse_characters_format(self, nxformat, data_type, char_matrix):
        """
        Given an XmlElement schema element ("format"), this parses the
        state definitions (if any) and characters (column definitions, if any),
        and populates the given char_matrix accordingly.
        """
        # if data_type == "standard":
        #     for nxstates in nxformat.findall_char_states():
        #         char_matrix.state_alphabets.append(self.parse_state_alphabet(nxstates))
        # else:
        #     pass
        if data_type in ("dna", "rna", "protein", "restriction"):
            # fixed alphabet: map to existing states
            for nxstates in nxformat.findall_char_states():
                state_alphabet_oid = nxstates.get("id", None)
                if state_alphabet_oid is not None:
                    self._id_state_alphabet_map[state_alphabet_oid] = char_matrix.default_state_alphabet
                for nxstate in nxstates.findall_char_state():
                    state_oid = nxstate.get('id', None)
                    label = nxstate.get('label', None)
                    symbol = nxstate.get('symbol', None)
                    token = nxstate.get('token', None)
                    try:
                        state = char_matrix.default_state_alphabet[symbol]
                    except KeyError:
                        raise Exception("'{}' is not a recognized symbol for the state alphabet for the '{}' data type".format(symbol, data_type))
                    assert (char_matrix.default_state_alphabet, state_oid) not in self._id_state_map
                    self._id_state_map[ (char_matrix.default_state_alphabet, state_oid) ] = state
                for nxstate in nxstates.findall_polymorphic_state_set():
                    state_oid = nxstate.get('id', None)
                    symbol = nxstate.get('symbol', None)
                    try:
                        state = char_matrix.default_state_alphabet[symbol]
                    except KeyError:
                        raise Exception("'{}' is not a recognized symbol for the state alphabet for the '{}' data type".format(symbol, data_type))
                    assert (char_matrix.default_state_alphabet, state_oid) not in self._id_state_map
                    self._id_state_map[ (char_matrix.default_state_alphabet, state_oid) ] = state
                for nxstate in nxstates.findall_uncertain_state_set():
                    state_oid = nxstate.get('id', None)
                    symbol = nxstate.get('symbol', None)
                    try:
                        state = char_matrix.default_state_alphabet[symbol]
                    except KeyError:
                        raise Exception("'{}' is not a recognized symbol for the state alphabet for the '{}' data type".format(symbol, data_type))
                    assert (char_matrix.default_state_alphabet, state_oid) not in self._id_state_map
                    self._id_state_map[ (char_matrix.default_state_alphabet, state_oid) ] = state
        else:
            for nxstates in nxformat.findall_char_states():
                char_matrix.state_alphabets.append(self.parse_state_alphabet(nxstates))
        for nxchars in nxformat.findall_char():
            col = char_matrix.new_character_type()
            char_state_set_id = nxchars.get('states')
            if char_state_set_id is not None:
                state_alphabet = self._id_state_alphabet_map.get(char_state_set_id, None)
                if state_alphabet is None:
                    raise Exception("State set '%s' not defined" % char_state_set_id)
                col.state_alphabet = state_alphabet
            elif hasattr(char_matrix, "default_state_alphabet") \
                and char_matrix.default_state_alphabet is not None:
                col.state_alphabet = char_matrix.default_state_alphabet
            char_matrix.character_types.append(col)
            chartype_id = nxchars.get('id')
            self._chartype_id_to_pos_map[chartype_id] = len(self._chartype_id_to_pos_map)
            self._id_chartype_map[chartype_id] = col
            self._char_types.append(col)
            annotations = [i for i in nxchars.findall_annotations()]
            for annotation in annotations:
                self._parse_annotations(col, annotation)

    def create_standard_character_alphabet(self, char_matrix, symbol_list=None):
        """
        Returns a standard character state alphabet based on symbol_list.
        Defaults to '0' - '9' if not specified.
        """
        if symbol_list is None:
            symbol_list = [str(i) for i in range(10)]
        state_alphabet = StateAlphabet()
        for s in symbol_list:
            state_alphabet.append(StateIdentity(symbol=s))
        char_matrix.state_alphabets.append(state_alphabet)
        char_matrix.default_state_alphabet = state_alphabet



#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Population genetic simlations.
"""

import random
import copy

from dendropy.utility import GLOBAL_RNG
from dendropy.interop import seqgen
from dendropy.model import discrete
from dendropy.model import coalescent
import dendropy

class FragmentedPopulations(object):

    def __init__(self,
                 div_time_gens,
                 num_desc_pops = 2,
                 mutrate_per_site_per_generation=10e-8,
                 desc_pop_size=10000,
                 use_seq_gen=False,
                 rng=GLOBAL_RNG):
        """
        __init__ arguments:

            - ``div_time_gens`` : generations since divergence,
            - ``num_desc_pops`` : number of descendent populations,
            - ``mutrate_per_site_per_generation`` : sequence mutation rate, per-site per-generation
            - ``desc_diploid_pop_size`` : descendent lineage population size (=N; ancestral pop size = num_desc_pops * N)
            - ``rng`` : random number generator
        """
        self.div_time_gens = div_time_gens
        self.num_desc_pops = num_desc_pops
        self.mutrate_per_site_per_generation = mutrate_per_site_per_generation
        self.desc_pop_size = desc_pop_size
        self.rng = rng
        self.kappa = 1.0
        self.base_freqs=[0.25, 0.25, 0.25, 0.25]
        self.seqgen_path = 'seq-gen'
        self.use_seq_gen = use_seq_gen
        self.gene_tree = None
        self.pop_tree = None
        self.mutation_tree = None

    def _get_theta(self):
        return 4 * self.mutrate_per_gene_per_generation * self.desc_pop_size

    def generate_sequences(self,
                           species_name,
                           samples_per_pop=10,
                           seq_len=2000,
                           use_seq_gen=True):

        self.generate_pop_tree(species_name=species_name, samples_per_pop=samples_per_pop)
        self.generate_gene_tree(species_name=species_name, samples_per_pop=samples_per_pop)
        d = dendropy.DataSet(self.mutation_tree.taxon_namespace)
        if self.use_seq_gen is True:
            sg = seqgen.SeqGen()
            sg.seqgen_path = self.seqgen_path
            sg.num_replicates = 1
            sg.quiet = True
            sg.rng = self.rng
            sg.seq_len = seq_len
            sg.char_model = 'HKY'
            sg.ti_tv = float(self.kappa) / 2
            sg.state_freqs = self.base_freqs
            sg.trees = [self.mutation_tree]
            d = sg.generate_dataset(dataset=d)
        else:
            char_matrix = discrete.hky85_chars(
                    seq_len=seq_len,
                    tree_model=self.mutation_tree,
                    mutation_rate=1.0,
                    kappa=1.0,
                    base_freqs=[0.25, 0.25, 0.25, 0.25],
                    root_states=None,
                    rng=self.rng)
            d.add_char_matrix(char_matrix)
        return d

    def generate_pop_tree(self, species_name, samples_per_pop=10):
        tree_data = { 'sp': species_name, 'divt': self.div_time_gens }
        desc_lineages = []
        for i in range(self.num_desc_pops):
            tree_data['id'] = i+1
            desc_lineages.append("%(sp)s%(id)d:%(divt)d" % tree_data)
        tree_string = "(" + (",".join(desc_lineages)) + ("):%d" % 0) #% (self.num_desc_pops * self.desc_pop_size * 10))
        self.pop_tree = dendropy.Tree.get_from_string(tree_string, schema="newick")
        return self.pop_tree

    def generate_gene_tree(self, species_name, samples_per_pop=10):
        """
        Given:
            ``species_name`` : string identifying species/taxon
            ``samples_per_pop`` : number of samples (genes) per population
        Returns:
            DendroPy tree, with branch lengths in generations
        """
        if self.pop_tree is None:
            self.generate_pop_tree(species_name, samples_per_pop=10)
        for idx, leaf in enumerate(self.pop_tree.leaf_iter()):
            if idx == 1:
                # ancestral population = num_desc_pops * desc population
                leaf.parent_node.edge.pop_size = self.num_desc_pops * self.desc_pop_size
            leaf.edge.pop_size = self.desc_pop_size
            leaf.num_genes = samples_per_pop
        self.gene_tree, self.pop_tree = coalescent.constrained_kingman_tree(self.pop_tree,
                                                          gene_node_label_fn=lambda x,y: "%sX%d" % (x,y),
                                                          rng=self.rng)

        self.mutation_tree = copy.deepcopy(self.gene_tree)
        for edge in self.mutation_tree.preorder_edge_iter():
            edge.length = edge.length * self.mutrate_per_site_per_generation
        return self.gene_tree

def pop_gen_tree(tree=None,
                 taxon_namespace=None,
                 ages=None,
                 num_genes=None,
                 pop_sizes=None,
                 num_genes_attr = 'num_genes',
                 pop_size_attr = 'pop_size',
                 rng=None):
    """
    This will simulate and return a tree with edges decorated with
    population sizes and leaf nodes decorated by the number of genes
    (samples or lineages) in each leaf.

    If ``tree`` is given, then this is used as the tree to be decorated.
    Otherwise, a Yule tree is generated based on the given taxon_namespace.
    Either ``tree`` or ``taxon_namespace`` must be given.

    The timing of the divergences can be controlled by specifying a vector of
    ages, ``ages``. This should be sequences of values specifying the ages of the
    first, second, third etc. divergence events, in terms of time from the
    present, specified either in generations (if the ``pop_sizes`` vector is
    given) or population units (if the pop_size vector is not given).
    If an ages vector is given and there are less than num_pops-1 of these,
    then an exception is raised.

    The number of gene lineages per population can be specified through
    the 'num_genes', which can either be an scalar integer or a list.
    If it is an integer, all the population get the same number of
    genes. If it is a list, it must be at least as long as num_pops.

    The population sizes of each edge can be specified using the ``pop_sizes``
    vector, which should be a sequence of values specifying the population
    sizes of the edges in postorder. If the pop_size vector is given, then it
    must be at least as long as there are branches on a tree, i.e. 2 * num_pops
    + 1, otherwise it is an error.  The population size should be the effective
    *haploid* population size; i.e., number of gene copies in the population: 2
    * N in a diploid population of N individuals, or N in a haploid population
    * of N individuals.

    If ``pop_size`` is 1 or 0 or None, then edge lengths of the tree are in
    haploid population units; i.e. where 1 unit of time equals 2N generations
    for a diploid population of size N, or N generations for a haploid
    population of size N. Otherwise edge lengths of the tree are in
    generations.

    This function first generates a tree using a pure-birth model with a
    uniform birth rate of 1.0. If an ages vector is given, it then sweeps
    through the internal nodes, assigning branch lengths such that the
    divergence events correspond to the ages in the vector. If a population
    sizes vector is given, it then visits all the edges in postorder, assigning
    population sizes to the attribute with the name specified in
    'pop_size_attr' (which is persisted as an annotation). During this, if an
    ages vector was *not* given, then the edge lengths are multiplied by the
    population size of the edge so the branch length units will be in
    generations. If an ages vector was given, then it is assumed that the ages
    are already in the proper scale/units.
    """

    # get our random number generator
    if rng is None:
        rng = GLOBAL_RNG # use the global rng by default

    # get a yule tree
    if not tree:
        if taxon_namespace:
            from dendropy.simulate import treesim
            tree = treesim.uniform_pure_birth_tree(taxon_namespace=taxon_namespace,
                                      rng=rng)
        else:
            raise Exception("Either tree or taxon namespace must be given")

    num_pops = len(tree.leaf_nodes())

    # basic idiot-checking
    if ages is not None and len(ages) < (num_pops - 1):
        msg = "Too few ages specified."
        raise Exception(msg)
    if num_genes is not None:
        if isinstance(num_genes, list):
            if len(num_genes) < num_pops:
                msg = "Too few number of gene samples specified"
                raise Exception(msg)
            else:
                samples = num_genes
        else:
            samples = [num_genes for tax in range(num_pops)]
    else:
        samples = None
    if pop_sizes is not None and len(pop_sizes) < (2 * num_pops + 1):
        msg = "Too few population sizes specified."
        raise Exception(msg)

    # set the ages
    if ages is not None:

        # get the internal nodes on the tree in reverse branching
        # order, so that newest nodes are returned first
        nodes = tree.nodes(filter_fn = lambda x : not x.is_leaf())
        nodes.sort(key=lambda x: x.distance_from_root())
        # assign the ages
        for index, node in enumerate(nodes):
            for child in node.child_nodes():
                child.edge.length = ages[index] - child.distance_from_tip()

    # set the gene samples
    if samples is not None:
        for index, leaf in enumerate(tree.leaf_node_iter()):
            setattr(leaf, num_genes_attr, samples[index])
            leaf.annotations.add_bound_attribute(num_genes_attr)

    # set the population sizes
    if pop_sizes is not None:
        index = 0
        for edge in tree.postorder_edge_iter():
            setattr(edge, pop_size_attr, pop_sizes[index])
            edge.annotations.add_bound_attribute(pop_size_attr)
            if ages is None:
                edge.length = edge.length * getattr(edge, pop_size_attr)
            index = index + 1

    return tree




#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Wrapper around calls to MUSCLE
"""

import dendropy
import subprocess
from dendropy.utility import processio

def muscle_align(char_matrix, muscle_args=None, muscle_path='muscle'):
    cmd = [muscle_path]
    if muscle_args:
        cmd = cmd + muscle_args
    p = subprocess.Popen(cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
    stdout, stderr = processio.communicate(p, char_matrix.as_string("fasta"))
    if p.returncode:
        raise Exception(stderr)
    d = char_matrix.__class__.get_from_string(stdout,
            "fasta",
            taxon_namespace=char_matrix.taxon_namespace)
    return d


